# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Created on Sat Feb  9 15:57:52 2019

@author: Nick Anthony
"""
from __future__ import annotations
import json
import logging
import os
import dataclasses
from enum import Enum, auto
from glob import glob
import h5py
import numpy as np
from scipy import io as spio
from shapely import geometry, wkb
import cv2
from rasterio import features
import shapely
import typing as t_
import copy
import pwspy.dataTypes._metadata as metadata


@dataclasses.dataclass(frozen=True)
class CameraCorrection:
    """This class represents all the information needed to correct camera related hardware defects in our data. This
    includes a dark count value (The counts registered when no light is incident on the camera. It also includes a
    polynomial that is used to linearize the counts. E.G. if you image something over a range of exposure times you would
    expect the measured counts to be proportional to the exposure time. However on some cameras this is not the case.

    Attributes:
        darkCounts: Dark count for a single pixel of the camera. This will be subtracted from the data in pre-processing.
            When binning is used the dark counts are summed together, so if you measure a dark count of 400 with 2x2
            binning then the single pixel dark count is 100.
        linearityPolynomial: Sequence of polynomial coefficients [a,b,c,etc...] in the order a*x + b*x^2 + c*x^3 + etc...
            Used to linearize the counts from the camera so that they are linearly proportional to the image brightness.
            This can generally be left as `None` for sCMOS cameras but it is often required for CCD type cameras.
    """
    darkCounts: float
    linearityPolynomial: t_.Sequence[float, ...] = None

    def __post_init__(self):
        # This code runs right after the built-in dataclass initializer runs.
        # Force the linearity polynomial to be a tuple.
        if self.linearityPolynomial is not None:
            object.__setattr__(self, 'linearityPolynomial', tuple(self.linearityPolynomial))
            assert isinstance(self.linearityPolynomial, tuple)

    def toJsonFile(self, filePath: str):
        """
        Save the camera correction to a JSON formatted text file.

        Args:
            filePath: The file path for the new JSON file.
        """
        if os.path.splitext(filePath)[-1] != '.json':
            filePath = filePath + '.json'
        with open(filePath, 'w') as f:
            json.dump(dataclasses.asdict(self), f)

    @classmethod
    def fromJsonFile(cls, filePath: str) -> CameraCorrection:
        """
        Load the camera correction from a json text file.

        Args:
             filePath: The file path of the JSON file to load from.

        Returns:
            A new instance of `CameraCorrection`.

        Examples:
            corr = CameraCorrection.fromJsonFile('~/Desktop/camera.json')

        """
        with open(filePath, 'r') as f:
            return cls(**json.load(f))


class Roi:
    """This class represents a single Roi used to select a specific region of an image. The Roi consists of a `mask` (a boolean array specifying which pixels are
    included in the Roi), a set of of `vertices` (a 2 x N array specifying the vertices of the polygon enclosing the
    mask, this is useful if you want to adjust the Roi later. Rather than calling the constructor directly you will
    generally create one of these objects through one of the `class methods` that construct one for you.

    Args:
        mask: A 2D boolean array where the True values indicate pixels that are within the ROI.
        verts: Can be a sequence of 2D (x, y) coordinates indicating the border of the ROI or a shapely `Polygon`.
            If an array of coordinates is used then it will be converted to the shell of a shapely polygon internally.
            While this information is partially redundant with the mask it is useful for many applications and can be
            complicated to calculate from `mask`.
    """

    def __init__(self, mask: np.ndarray, verts: t_.Union[np.ndarray, geometry.Polygon]):
        assert isinstance(mask, np.ndarray), f"Mask data is of type: {type(mask)}. Must be numpy array."
        assert len(mask.shape) == 2
        assert mask.dtype == np.bool
        self.polygon: geometry.Polygon
        if isinstance(verts, geometry.MultiPolygon):  # I'm not sure how but it is possible to get a multipolygon. In this case just select out the biggest polygon.
            verts = verts[0]
        if isinstance(verts, geometry.Polygon):
            self.polygon = verts
        else:
            assert len(verts.shape) == 2
            assert verts.shape[1] == 2
            self.polygon = geometry.Polygon(shell=verts)
        self.polygon = self.polygon.buffer(0)  # This little trick `normalizes` the format of the polygon so that holes will plot properly. https://gis.stackexchange.com/questions/374001/plotting-shapely-polygon-with-holes-does-not-plot-all-holes

        self.mask = mask

    @property
    def verts(self) -> np.ndarray:
        """An array of vertices for the outer ring of the polygon. For most ROIs they only have an outer ring anyway."""
        return np.array(self.polygon.exterior.coords)

    @classmethod
    def fromVerts(cls, verts: np.ndarray, dataShape: t_.Tuple[float, float]) -> Roi:
        """
        Automatically generate the mask for an Roi using just the vertices of an enclosing polygon.

        Args:
            verts: A sequence of 2D (x, y) coordinates indicating the border of the ROI.
            dataShape: A tuple giving the shape of the array that this Roi is associated with.

        Returns:
            A new instance of `Roi`

        Examples:
            myRoi = Roi.fromVerts('nucleus', 1, polyVerts, (1024, 1024))

        """
        assert isinstance(verts, np.ndarray)
        assert isinstance(dataShape, tuple)
        assert len(dataShape) == 2
        assert verts.shape[1] == 2
        assert len(verts.shape) == 2
        iVerts = np.rint([verts]).astype(np.int32)  # The brackets here convert to a 3d array which is what cv2.fillpoly expects. We have to round to integers for cv2 to work.
        mask = np.zeros(dataShape, dtype=np.int32)
        cv2.fillPoly(mask, iVerts, 1)
        mask = mask.astype(bool)
        return cls(mask, verts)

    @classmethod
    def fromMask(cls, mask: np.ndarray) -> Roi:
        """
        Use rasterio to create find the vertices of a mask.
        Args:
            mask: A boolean array. The mask have only one contiguous `True` region

        Returns:
            A new instance of `Roi`

        TODO:
            This function doesn't work properly if there is a `False` region of `mask` completely enclosed by a `True` region of `mask`.
        """

        all_polygons = []
        for shape, value in features.shapes(mask.astype(np.uint8), mask=mask):
            all_polygons.append(shapely.geometry.shape(shape))

        poly = sorted(all_polygons, key=lambda ply: ply.area)[-1]  # Return the biggest found polygon
        return cls(mask=mask, verts=poly)

    def transform(self, matrix: np.ndarray) -> Roi:
        """Return a copy of this Roi that has been transformed by an affine transform matrix like the one returned by
        opencv.estimateRigidTransform. This can be obtained using the functions in the utility.machineVision module.

        Args:
            matrix: A 2x3 numpy array representing an affine transformation.
        Returns:
            A new instance of Roi representing this Roi after transformation.
        """
        mask = cv2.warpAffine(self.mask.astype(np.uint8), matrix, self.mask.shape).astype(np.bool)
        verts = cv2.transform(self.verts[None, :, :], matrix)[0, :, :]  # For some reason this needs to be 3d for opencv to work.
        return Roi(mask=mask, verts=verts)


class RoiFile:
    """This class represents a single Roi File used to save and load an ROI. Each Roi File is identified by a
    `name` and a `number`. The recommended file format is HDF2, in this format multiple rois of the same name but differing
    numbers can be saved in a single HDF file.

    Args:
        name: The name used to identify this ROI. Multiple ROIs can share the same name but must have unique numbers.
        number: The number used to identify this ROI. Each ROI with the same name must have a unique number.
        roi: The ROI object associated with this file.
        filePath: The path to the file that this object was loaded from.
        fileFormat: The format of the file that this object was loaded from.
        acquisition: The acquisition object that this ROI belongs to.
    """

    class FileFormats(Enum):
        """An enumerator of the different file formats that an ROI can be saved to."""
        MAT = auto()  # The oldest file format. Each ROI was saved to its own matlab .mat file as a boolean mask.
        HDF = auto()  # This was originally the default file format of this Python software. Each ROI of the same name was saved as a dataset in an HDF file. The dataset contained the boolean mask.
        HDF2 = auto()  # For a long time this was the default. Each ROI of the same name is saved as an H5PY.Group in an HDF file. Each ROI group contains a dataset for the boolean mask as well as a dataset for the XY coordinates of the enclosing polygon. This saves us from having to constantly recalculate the outline of the ROI for processing purposes.
        HDF3 = auto()  # On 3/26/2021 We switched to this from HDF2. Rather than verts we now store 'wkb' of the underlying shapely file. Allow for ROIs with holes, and other more complex situations.

    def __init__(self, name: str, number: int, roi: Roi, filePath: str, fileFormat: RoiFile.FileFormats, acquisition: metadata.Acquisition):
        self._roi = roi
        self.name = name
        self.number = number
        self.filePath = filePath
        self.fformat = fileFormat
        self.acquisition = acquisition
        self.isDeleted = False

    def getRoi(self) -> Roi:
        """Return the ROI object associated with this file.

        Returns:
            The `Roi` object containing geometry information.
        """
        return copy.deepcopy(self._roi)  # Rois are mutable so return a copy.

    def __repr__(self):
        return f"RoiFile({self.name}, {self.number})"

    @staticmethod
    def getValidRoisInPath(path: str) -> t_.List[t_.Tuple[str, int, RoiFile.FileFormats]]:
        """Search the `path` for valid roiFile files and return the detected rois as a list of tuple where each tuple
        contains the `name`, `number`, and file format for the Roi.

        Args:
            path: The path to the folder containing the Roi files.

        Returns:
            A list of tuples containing:
                name: The detected Roi name
                number: The detected Roi number
                fformat: The file format of the file that the Roi is stored in
        """
        patterns = [('BW*_*.mat', RoiFile.FileFormats.MAT), ('ROI_*.h5', RoiFile.FileFormats.HDF)]
        files = {fformat: glob(os.path.join(path, p)) for p, fformat in patterns}  # Lists of the found files keyed by file format
        ret = []
        for fformat, fileNames in files.items():
            if fformat == RoiFile.FileFormats.HDF:  # Could still technically be HDF2 or HDF3
                for i in fileNames:
                    with h5py.File(i, 'r') as hf:  # making sure to open this file in read mode makes the function way faster!
                        for g in hf.keys():
                            if isinstance(hf[g], h5py.Group):  # HDF3 or HDF2 file format
                                if 'fileFormat' in hf[g].attrs:  # HDF3 format
                                    assert 'wkb' in hf[g]
                                    assert 'ROI_' in i
                                    name = i.split("ROI_")[-1][:-3]
                                    try:
                                        ret.append((name, int(g), RoiFile.FileFormats.HDF3))
                                    except ValueError:
                                        logging.getLogger(__name__).warning(f"File {i} contains uninterpretable dataset named {g}")
                                else:  # HDF2 did not have this fileformat attribute
                                    if 'mask' in hf[g] and 'verts' in hf[g]:
                                        assert 'ROI_' in i
                                        name = i.split("ROI_")[-1][:-3]
                                        try:
                                            ret.append((name, int(g), RoiFile.FileFormats.HDF2))
                                        except ValueError:
                                            logging.getLogger(__name__).warning(f"File {i} contains uninterpretable dataset named {g}")
                                    else:
                                        raise ValueError("File is missing datasets")
                            elif isinstance(hf[g], h5py.Dataset):  # Legacy format
                                assert 'roi_' in i
                                name = i.split('roi_')[-1][:-3]  # Old files used lower case rather than "ROI_"
                                try:
                                    ret.append((name, int(g), RoiFile.FileFormats.HDF))
                                except ValueError:
                                    logging.getLogger(__name__).warning(f"File {i} contains uninterpretable dataset named {g}")

            elif fformat == RoiFile.FileFormats.MAT:
                for i in fileNames:  # list in files
                    i = os.path.split(i)[-1]
                    if len(i.split("_")) != 2:  # Some old data has files that are not ROIs but are named almost identically, this helps us avoid bugs with them.
                        continue
                    num = int(i.split('_')[0][2:])
                    name = i.split('_')[1][:-4]
                    ret.append((name, num, RoiFile.FileFormats.MAT))
        return ret

    @staticmethod
    def deleteRoi(directory: str, name: str, number: int, fformat: t_.Optional[RoiFile.FileFormats] = None):
        """Delete the dataset associated with the Roi object specified by `name` and `num`.

        Args:
            directory: The path to the folder containing the Roi file.
            name: The name used to identify this ROI.
            number: The number used to identify this ROI.
            fformat: The format of the file.

        Raises:
            FileNotFoundError: If the file isn't found.
        """
        assert os.path.isdir(directory)
        if fformat is RoiFile.FileFormats.MAT:
            path = os.path.join(directory, f"BW{number}_{name}.mat")
        elif fformat in [RoiFile.FileFormats.HDF, RoiFile.FileFormats.HDF2, RoiFile.FileFormats.HDF3]:
            path = os.path.join(directory, f"ROI_{name}.h5")
        elif fformat is None:  # AutoDetect the file format
            try:
                RoiFile.deleteRoi(directory, name, number, fformat=RoiFile.FileFormats.MAT)
                return
            except FileNotFoundError:
                try:
                    RoiFile.deleteRoi(directory, name, number, fformat=RoiFile.FileFormats.HDF)
                    return
                except FileNotFoundError as e:
                    raise e
        else:
            raise Exception(f"fformat of {fformat} is not accepted")

        if not os.path.exists(path):
            raise FileNotFoundError(f"The ROI file {name},{number} and format {fformat} was not found in {directory}.")

        if fformat in [RoiFile.FileFormats.HDF, RoiFile.FileFormats.HDF2, RoiFile.FileFormats.HDF3]:
            with h5py.File(path, 'a') as hf:
                if np.string_(str(number)) not in hf.keys():
                    raise ValueError(f"The file {path} does not contain ROI number {number}.")
                del hf[np.string_(str(number))]
                remaining = len(list(hf.keys()))
            if remaining == 0:  # If the file is empty then remove it.
                os.remove(path)
        elif fformat is RoiFile.FileFormats.MAT:
            os.remove(path)
        else:
            raise Exception("Programming error.")

    @classmethod
    def fromHDF_legacy_legacy(cls, directory: str, name: str, number: int, acquisition: metadata.Acquisition = None) -> RoiFile:
        """Load an Roi from an older version of the HDF file format which did not include the vertices parameter.

        Args:
            directory: The path to the directory containing the HDF file.
            name: The name used to identify this ROI.
            number: The number used to identify this ROI.
        Raises:
            OSError: If the file was not found
        Returns:
            A new instance of Roi loaded from file
        """
        path = os.path.join(directory, f'ROI_{name}.h5')
        if not os.path.exists(path):
            raise OSError(f"File {path} does not exist.")
        with h5py.File(path, 'r') as hf:
            roi = Roi.fromMask(np.array(hf[str(number)]).astype(np.bool))
            return cls(name, number, roi, filePath=path, fileFormat=RoiFile.FileFormats.HDF, acquisition=acquisition)

    @classmethod
    def fromHDF_legacy(cls, directory: str, name: str, number: int, acquisition: metadata.Acquisition = None) -> RoiFile:
        """Load an Roi from an HDF file. Uses the old HDF2 format.

        Args:
            directory: The path to the directory containing the HDF file.
            name: The name used to identify this ROI.
            number: The number used to identify this ROI.
        Raises:
            OSError: If the file was not found
        Returns:
            A new instance of Roi loaded from file
        Examples:
            myRoi = Roi.fromHDF('~/Desktop', 'nucleus', 1)
        """
        path = os.path.join(directory, f'ROI_{name}.h5')
        if not os.path.exists(path):
            raise OSError(f"File {path} does not exist.")
        with h5py.File(path, 'r') as hf:
            dset = hf[str(number)]
            verts = dset['verts']
            if verts.shape is None:
                roi = Roi.fromMask(np.array(dset['mask']).astype(np.bool))  # Some old files could be saved without verts. allow loading them.
            else:
                roi = Roi(np.array(dset['mask']).astype(np.bool), verts=np.array(verts))
            return cls(name, number, roi, filePath=path, fileFormat=RoiFile.FileFormats.HDF2, acquisition=acquisition)

    @classmethod
    def fromHDF(cls, directory: str, name: str, number: int, acquisition: metadata.Acquisition = None) -> RoiFile:
        """Load an Roi from the newest ROI format of HDF file.

        Args:
            directory: The path to the directory containing the HDF file.
            name: The name used to identify this ROI.
            number: The number used to identify this ROI.
        Raises:
            OSError: If the file was not found
        Returns:
            A new instance of Roi loaded from file
        Examples:
            myRoi = Roi.fromHDF('~/Desktop', 'nucleus', 1)"""
        path = os.path.join(directory, f'ROI_{name}.h5')
        if not os.path.exists(path):
            raise OSError(f"File {path} does not exist.")
        with h5py.File(path, 'r') as hf:
            group = hf[str(number)]
            assert 'fileFormat' in group.attrs, "No fileFormat attribute found for the ROI file. Try using one of the legacy ROI loading methods."
            assert group.attrs['fileFormat'] == RoiFile.FileFormats.HDF3.name, f'Only HDF3 format is supported by this loading method, not {group.attrs["fileFormat"]}'
            wkbBytes = bytes(group['wkb'][()])
            polygon = wkb.loads(wkbBytes)
            mask = np.array(group['mask']).astype(np.bool)
            roi = Roi(mask, verts=polygon)
            return cls(name, number, roi, filePath=path, fileFormat=RoiFile.FileFormats.HDF3, acquisition=acquisition)

    @classmethod
    def fromMat(cls, directory: str, name: str, number: int, acquisition: metadata.Acquisition = None) -> RoiFile:
        """Load an Roi from a .mat file saved in matlab. This file format is not recommended as it does not include the
        `vertices` parameter which is useful for visually rendering and readjusting the Roi.

        Args:
            directory: The path to the directory containing the HDF file.
            name: The name used to identify this ROI.
            number: The number used to identify this ROI.
        Returns:
            A new instance of Roi loaded from file
        """
        filePath = os.path.join(directory, f'BW{number}_{name}.mat')
        spFile = spio.loadmat(filePath)
        if 'BW' in spFile.keys():
            mask = spFile['BW'].astype(np.bool)
        elif 'mask' in spFile.keys():
            mask = spFile['mask'].astype(np.bool)
        else:
            raise KeyError(f"A `mask` was not found in the `mat` file: {filePath}")
        roi = Roi.fromMask(mask)
        return cls(name, number, roi, filePath=filePath, fileFormat=RoiFile.FileFormats.MAT, acquisition=acquisition)

    @classmethod
    def loadAny(cls, directory: str, name: str, number: int, acquisition: metadata.Acquisition = None) -> RoiFile:
        """Attempt loading any of the known file formats.

        Args:
            directory: The path to the directory containing the HDF file.
            name: The name used to identify this ROI.
            number: The number used to identify this ROI.
        Returns:
            A new instance of Roi loaded from file
        """
        try:
            return RoiFile.fromHDF(directory, name, number, acquisition=acquisition)
        except:
            try:
                return RoiFile.fromHDF_legacy(directory, name, number, acquisition=acquisition)
            except:
                try:
                    return RoiFile.fromHDF_legacy_legacy(directory, name, number, acquisition=acquisition)
                except OSError:  # For backwards compatibility purposes
                    return RoiFile.fromMat(directory, name, number, acquisition=acquisition)

    @classmethod
    def toHDF(cls, roi: Roi, name: str, number: int, directory: str, overwrite: t_.Optional[bool] = False, acquisition: metadata.Acquisition = None) -> RoiFile:
        """
        Save the Roi to an HDF file in the specified directory. The filename is automatically chosen based on the
        `name` parameter of the Roi. Multiple Roi's with the same `name` will be saved into the same file if they have
        differing `number` parameters. If `overwrite` is true then any existing dataset will be replaced, otherwise an
        error will be raised.

        Args:
            roi: The ROI to save.
            name: The name name to save as. This will be part of the file name
            number: The ROI number to save as. Multiple ROIS of the same name can be saved to the same file but the numbers must be unique
            directory: The path of the folder to save the new HDF file to. The file will be named automatically based
                on the `name` attribute of the Roi
            overwrite: If True then if an Roi with the same `number` as this Roi is found it will be overwritten.
        """
        savePath = os.path.join(directory, f'ROI_{name}.h5')
        numStr = np.string_(str(number))
        mask = roi.mask.astype(np.uint8)
        with h5py.File(savePath, 'a') as hf:
            if numStr in hf.keys():
                if overwrite:
                    del hf[numStr]
                else:
                    raise OSError(f"The Roi file {savePath} already contains a dataset {number}")
            g = hf.create_group(numStr)
            g.attrs['fileFormat'] = RoiFile.FileFormats.HDF3.name
            g.create_dataset(np.string_("wkb"), data=np.void(roi.polygon.wkb))  # np.void is required here so we can save a byte array with `null` in it.
            g.create_dataset(np.string_("mask"), data=mask, compression=5)
        return cls(name, number, roi, filePath=savePath, fileFormat=RoiFile.FileFormats.HDF2, acquisition=acquisition)

    def delete(self):
        """
        Delete the dataset associated with the Roi object.

        """
        self.deleteRoi(os.path.split(self.filePath)[0], self.name, self.number)
        self.isDeleted = True

    def update(self, roi: Roi):
        """
        Save a new roiFile to the existing file.
        Args:
            roi: The updated ROI to save
        """
        assert not self.isDeleted
        if self.fformat not in [RoiFile.FileFormats.HDF2, RoiFile.FileFormats.HDF3]:
            raise NotImplementedError(f"RoiFile of format: {self.fformat} cannot be updated.")
        self.toHDF(roi, self.name, self.number, os.path.split(self.filePath)[0], overwrite=True)
        self._roi = copy.deepcopy(roi)  # We don't wont to use the same object that might still have external mutable references



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Created on Fri Apr  3 16:24:11 2020

@author: backman05
"""
import typing
from time import time

import numpy as np
import skan
import matplotlib.pyplot as plt
import pandas as pd
import skimage.morphology as morph
from scipy.ndimage import binary_hit_or_miss

from pwspy.utility.misc import cached_property


class Skel(skan.Skeleton):
    """Extends skan.Skeleton to allow for removing branches. Also conveniently packages the `summary` dataframe along with the skeleton.
    Can analyze a 2d or 3d array. Represents the full array which may contain multiple independent skeletons."""
    def __init__(self, img):
        super().__init__(img)
        self.summary = skan.summarize(self)
        self.summary['branch-type'] = self.summary['branch-type'].astype(int)
        self.summary['ids'] = self.summary.apply(lambda row: tuple(sorted((row['node-id-src'], row['node-id-dst']))), axis=1)

    def path_type(self, i):
        return self.summary.loc[i]['branch-type']

    def path_length(self, i):
        return self.path_lengths()[i]

    def remove_path(self, idxes: typing.List[int]):
        """Strip the paths assotiated with the indexes in idxes"""
        if isinstance(idxes, int):
            idxes = [idxes]
        idxes = set(idxes) #Make sure the indexes are unique
        if len(idxes) == self.n_paths:
            return None
        Coords = ([], [])
        for idx in idxes:
            indices = self.path(idx)
            ep1Deg = self.degrees[indices[0]] #Get the degrees (connectedness) of both endpoints
            ep2Deg = self.degrees[indices[-1]]
            if ep1Deg < 3:
                start = 0
            else:
                start = 1 #If this endpoint is shared with other paths don't delete it (It's coordinates are likly not integers anyway due to Skan's handling of clustered ponits at junctions)
            if ep2Deg < 3:
                end = None
            else:
                end = -1
            coords = self.path_coordinates(idx)
            coords = coords[start:end].astype(int)
            Coords = (Coords[0] + list(coords[:, 0]), Coords[1] + list(coords[:, 1]))
        img = self.skeleton_image
        img[Coords] = 0 # Set all the pixels associated with a deleted path to 0.
        #Remove isolated pixels
        struct = np.array([[0,0,0],[0,1,0],[0,0,0]], dtype=bool)
        isoPixels = binary_hit_or_miss(img, struct)
        img[isoPixels]=False
        img = morph.skeletonize(img) #Without this step we get hard crashes when creating a new Skel
        s = Skel(img)  # Re create the skeleton object.
        return s

    def plot(self):
        ndim = self.path_coordinates(1).shape[1]
        if ndim == 2:
            a = AnnotFig(self)
        elif ndim == 3:
            from mpl_toolkits.mplot3d import Axes3D
            import matplotlib as mpl
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            groups = self.summary.groupby('skeleton-id')
            colors = mpl.cm.get_cmap('gist_rainbow')(np.linspace(0, 1, num=len(groups)))
            for i, (n, group) in enumerate(groups):
                for idx in group.index:
                    coords = self.path_coordinates(idx)
                    coords = (coords[:,0], coords[:,1], coords[:,2])
                    ax.plot(*coords, c=colors[i])
            fig.show()
        else:
            raise ValueError(f"Not sure what to do with a {ndim} dimensional skeleton")

class SingleSkeleton:
    """Represents a single connect skeleton from the Skel class.
    This class is not used. the idea was to reject whole skeletons
    as the first step based on their geometry."""
    def __init__(self, parent: Skel, id: int):
        self.s = parent
        self.id = id
        from shapely.geometry import MultiLineString
        linesCoords = []
        for i in self.indices:
            c = self.s.path_coordinates(i)
            c = tuple(tuple(c[i, :]) for i in range(c.shape[0]))
            linesCoords.append(c)
        self.mls = MultiLineString(linesCoords)

    @cached_property
    def indices(self):
        return list(self.s.summary[self.s.summary['skeleton-id'] == self.id].index)

    @property
    def convexHull(self):
        return self.mls.convex_hull.area

    @property
    def length(self):
        return self.mls.length


class AnnotFig:
    """Pass a Skel to this class and it will plot all the paths with annotation giving the type and length."""
    def __init__(self, skel: Skel, title=''):
        self.fig, self.ax = plt.subplots()
        self.fig.annotfigref = self
        self.fig.suptitle(title)
        self.annot = self.ax.annotate("", xy=(0, 0), xytext=(20, 20), textcoords="offset points",
                                      bbox=dict(boxstyle="round", fc="w"),
                                      arrowprops=dict(arrowstyle="->"))

        self.lines = self.plotLines(skel)

        self.skel = skel

        cid = self.fig.canvas.mpl_connect('motion_notify_event', self._hoverCallback)

    def _hoverCallback(self, event):  # Show an annotation about the ROI when the mouse hovers over it.
        def update_annot(line, idx):
            self.annot.xy = line.get_xydata()[0, :]
            sid = self.skel.summary.iloc[idx]['skeleton-id']
            ss = SingleSkeleton(self.skel, sid)
            text = f"type:{self.skel.path_type(idx)}, length:{self.skel.path_length(idx):.1f}, skel_length:{ss.length},skel_rat:{ss.convexHull/ss.length}"
            self.annot.set_text(text)
            self.annot.get_bbox_patch().set_alpha(0.4)

        vis = self.annot.get_visible()
        if event.inaxes == self.ax:
            for line, index in self.lines:
                contained, _ = line.contains(event)
                if contained:
                    if not vis:
                        update_annot(line, index)
                        self.annot.set_visible(True)
                        self.fig.canvas.draw_idle()
                    return
            if vis:  # If we got here then no hover actions were found.
                self.annot.set_visible(False)
                self.fig.canvas.draw_idle()

    def plotLines(self, skeleton):
        lines = []
        for n in range(skeleton.n_paths):
            coords = skeleton.path_coordinates(n)
            coords = (coords[:, 0], coords[:, 1])
            line = self.ax.plot(*coords)[0]
            lines.append((line, n))
        return lines

    def show(self):
        self.fig.show()


def _scanIds(s: Skel, ids: typing.List[int]):
    """Returns a list of indexes of skeleton paths to be removed"""
    idxes = []
    for i in ids:
        matchArr = np.any(s.summary[['node-id-src', 'node-id-dst']] == i, axis=1) #Find all row with an endpoint matching `i`. this is kinda slow. this is the main place that would benefit from optimization
        a = s.summary[matchArr]  # Select all rows using this endpoint
        if len(a) > 1:  # multiple branches use this endpoint.
            a = a[a['branch-type'] == 1]
            if len(a) == 0:
                continue
            idx = a['branch-distance'].idxmin()  # delete the shortest one
            if a['branch-distance'].min() < 50:
                idxes.append(idx)
        elif len(a) == 1:
            if a['branch-type'].iloc[0] == 0:  # This is an isolated segment
                if a['branch-distance'].iloc[0] < 50:  # short Stray segment
                    idxes.append(a.index[0])
        else:
            print(f"No paths found for id {i}")  # This shouldn't happen
    return idxes

def _scanSharedEndpoints(s):
    """Returns a list of indexes of skeleton paths to be removed"""
    g = s.summary.groupby('ids')
    idxes = []
    for i, group in g:  # Groups of paths that share both end points. These could be loops or groups of branches that start and stop in the same place
        loops = group[group['branch-type'] == 3]
        loopIdxes = list(loops[loops['branch-distance'] <= 40].index)  # We're going to discard all loops that are too short.

        juncs = group[group['branch-type'] == 2]  # Junction-junction branches
        juncs = juncs.sort_values('branch-distance')
        twinIdxes = list(juncs.index)[1:]  # We're going to discard all paths except for the shortest path.
        idxes += loopIdxes + twinIdxes
    return idxes

# The types are: - tip-tip (0) - tip-junction (1) - junction-junction (2) - path-path (3)
# TODO a curviness parameter based on distance/vs euclidian distance

def prune2dIm(s):
    """Take Skel, s, and recursively prune unwanted paths."""
    iteration = 0
    while True:
        print(f"Iteration {iteration + 1}")
        iteration += 1

        #Generate all the unique IDs
        ids = pd.concat([s.summary['node-id-src'], s.summary['node-id-dst']]).drop_duplicates()
        print(f"IDs to scan: {len(ids)}")
        idxes = _scanIds(s, ids)

        print(f"Found Step1 {len(idxes)}")
        if len(idxes) > 0:
            s = s.remove_path(idxes)
            if s is None:  # All paths were removed, no more recursion needed
                break
            else:
                continue

        idxes = _scanSharedEndpoints(s)

        print(f"Found Step2 {len(idxes)}")
        if len(idxes) > 0:
            s = s.remove_path(idxes)
            if s is None:  # All paths were removed, no more recursion needed
                break
            else:
                continue

        break #We only get to this point if we didn't take any action above. We are, therefore, done.
    return s




def prune3dIm(skelIm, plotEveryN=3):
    """Loop through the planes of a 3d array and prune each 2d plane using prune2dIm"""
    skel = skelIm.copy()
    out = np.zeros_like(skel)

    try:
        for z in range(skelIm.shape[2]):
            print(F"Z={z}")
            if np.sum(skel[:, :, z]) == 0:
                continue
            s = Skel(skel[:, :, z].copy())

            if z % plotEveryN == 0:
                beforeIm = AnnotFig(s, title=f'Before pruning {z}')
                beforeIm.show()

            s = prune2dIm(s)

            if s is None:
                continue  #:Leave this plane of the image as zeros since the whole skeleton was removed
            else:
                if z % plotEveryN == 0:
                    afterIm = AnnotFig(s, title=f'After pruning {z}')
                    afterIm.show()
                out[:, :, z] = s.skeleton_image
    except KeyboardInterrupt: # Allow early stopping though keybord interrupt
        pass
    return out


def _concaveHull3d(coords: typing.List[typing.Tuple[int, int]], alpha):
    """
    Found here: https://gist.github.com/dwyerk/10561690
    Compute the alpha shape (concave hull) of a set
    of points.

    Args:
        coords: nx2 array of points.
        alpha: alpha value to influence the
            gooeyness of the border. Smaller numbers
            don't fall inward as much as larger numbers.
            Too large, and you lose everything!
    """
    from scipy.spatial.qhull import Delaunay
    from shapely import geometry
    from shapely.ops import cascaded_union, polygonize
    if len(coords) < 4:
        # When you have a triangle, there is no sense
        # in computing an alpha shape.
        return geometry.MultiPoint(coords).convex_hull
    coords = np.array(coords)
    tri = Delaunay(coords)
    triangles = coords[tri.simplices]
    # Lengths of sides of triangle
    a = ((triangles[:, 0, 0] - triangles[:, 1, 0]) ** 2 + (triangles[:, 0, 1] - triangles[:, 1, 1]) ** 2) ** 0.5
    b = ((triangles[:, 1, 0] - triangles[:, 2, 0]) ** 2 + (triangles[:, 1, 1] - triangles[:, 2, 1]) ** 2) ** 0.5
    c = ((triangles[:, 2, 0] - triangles[:, 0, 0]) ** 2 + (triangles[:, 2, 1] - triangles[:, 0, 1]) ** 2) ** 0.5
    s = (a + b + c) / 2.0  # Semiperimeter of triangle

    areas = (s * (s - a) * (s - b) * (s - c)) ** 0.5  # Area of triangle by Heron's formula

    circums = a * b * c / (4.0 * areas)
    filtered = triangles[circums < alpha]  # Here's the radius filter.

    edge1 = filtered[:, (0, 1)]
    edge2 = filtered[:, (1, 2)]
    edge3 = filtered[:, (2, 0)]
    edge_points = np.unique(np.concatenate((edge1, edge2, edge3)), axis=0).tolist()
    m = geometry.MultiLineString(edge_points)
    triangles = list(polygonize(m))
    return cascaded_union(triangles), edge_points


if __name__ == '__main__':
    skel = np.load(r'C:\Users\backman05\Desktop\findsurface\skel.npy')
    s = Skel(skel[:,:,13])
    stime = time()
    s = prune2dIm(s)
    print("HHAHAHA", time()-stime)
    out = prune3dIm(skel)


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import json
import logging
import multiprocessing as mp
import os
import pathlib
import subprocess
import sys
import typing as t_
import abc
import warnings
from datetime import datetime
import enum
import h5py
import jsonschema
import numpy as np
import tifffile as tf
from scipy import io as spio
from pwspy.dataTypes import _jsonSchemasPath
from pwspy.dataTypes._other import CameraCorrection, Roi, RoiFile
import pwspy.dataTypes._data as pwsdtd
from pwspy import dateTimeFormat
from pwspy.utility.misc import cached_property
if t_.TYPE_CHECKING:
    from pwspy.analysis import AbstractHDFAnalysisResults


class MetaDataBase(abc.ABC):
    """
    This base class provides that basic functionality to store information about a PWS related acquisition on file.

    Args:
        metadata: A dictionary containing the metadata
        filePath: The path to the location the metadata was loaded from
        acquisitionDirectory: A reference to the `Acquisition` associated with this object.

    """

    @property
    @abc.abstractmethod
    def _jsonSchemaPath(self) -> str:
        """Each sublass should provide a path to the jsonschema file. this is needed in order to resolve jsonschema references"""
        pass

    @property
    @abc.abstractmethod
    def _jsonSchema(self) -> dict:
        """Each subclass should provide a json schema loaded from a file.
        This serves as a schematic that can be checked against when loading metadata to make sure it contains the required information."""
        pass

    def __init__(self, metadata: dict, filePath: t_.Optional[str] = None, acquisitionDirectory: t_.Optional[Acquisition] = None):
        logger = logging.getLogger(__name__)
        self.filePath = filePath
        self.acquisitionDirectory = acquisitionDirectory
        refResolver = jsonschema.RefResolver(pathlib.Path(self._jsonSchemaPath).as_uri(), None)  # This resolver is used to allow derived json schemas to refer to the base schema.
        jsonschema.validate(instance=metadata, schema=self._jsonSchema, resolver=refResolver, cls=_TupleValidator)
        self.dict: dict = metadata
        try:
            datetime.strptime(self.dict['time'], dateTimeFormat)
        except ValueError:
            try:
                logger.info("Detected a non-compliant timestamp. attempting to correct.")
                self.dict['time'] = datetime.strftime(datetime.strptime(self.dict['time'], "%d-%m-%y %H:%M:%S"), dateTimeFormat)
            except ValueError:
                logger.warning("The time stamp could not be parsed. Replacing with 1_1_1970")
                self.dict['time'] = "1-1-1990 01:01:01"
        if self.dict['system'] == "":
            logger.warning("The `system` name in the metadata is blank. Check that the PWS System is saving the proper calibration values.")
        if all([i in self.dict for i in ['darkCounts', 'linearityPoly']]):
            if self.dict['darkCounts'] == 0:
                logger.warning("Detected a darkCounts value of 0 in the pwsdtd.PwsCube Metadata. Check that the PWS System is saving the proper calibration values.")
            self.cameraCorrection = CameraCorrection(darkCounts=self.dict['darkCounts'],
                                                     linearityPolynomial=self.dict['linearityPoly'])
        else:
            self.cameraCorrection = None

    @abc.abstractmethod
    def toDataClass(self, lock: t_.Optional[mp.Lock]) -> pwsdtd.ICBase:
        """Convert the metadata class to a class that loads the data

        Args:
            lock: A `Lock` object used to synchronize IO in multithreaded and multiprocessing applications.
        """
        pass

    @property
    @abc.abstractmethod
    def idTag(self) -> str:
        """A string that uniquely identifies this data."""
        pass

    @property
    def binning(self) -> int:
        """
        The binning setting used by the camera. This is needed in order to properly correct dark counts.
        This is generally extracted from metadata saved by Micromanager
        """
        return self.dict['binning']

    @property
    def pixelSizeUm(self) -> float:
        """
        The pixelSize expressed in microns. This represents the length of each square pixel in object space. Binning
        has already been accounted for here. This is generally extracted from metadata saved my MicroManager
        """
        return self.dict['pixelSizeUm']

    @property
    def exposure(self) -> float:
        """The exposure time of the camera expressed in milliseconds."""
        return self.dict['exposure']

    @property
    def time(self) -> str:
        """The date and time that the acquisition was taken."""
        return self.dict['time']

    @property
    def systemName(self) -> str:
        """The name of the system this was acquired on. The name is set in the `PWS Acquisition Plugin` for Micromanager."""
        return self.dict['system']

    @staticmethod
    def decodeHdfMetadata(d: h5py.Dataset) -> dict:
        """Attempt to extract a dictionary of metadata from an HDF5 dataset.

        Args:
            d: The `h5py.Dataset` to load from.
        Returns:
            A dictionary containing the metadata
        """
        assert 'metadata' in d.attrs
        return json.loads(d.attrs['metadata'])

    def encodeHdfMetadata(self, d: h5py.Dataset) -> h5py.Dataset:
        """Save this metadata object as a json string in an HDF5 dataset.

        Args:
            d: The `h5py.Dataset` to save the metadata to.
        """
        d.attrs['metadata'] = np.string_(json.dumps(self.dict))
        return d


class AnalysisManager(abc.ABC):
    """Handles the functionality to save, load, etc. analysis files.

    Args:
        metadata: A dictionary containing the metadata
        filePath: The path to the location the metadata was loaded from
        acquisitionDirectory: A reference to the `Acquisition` associated with this object.

    """
    def __init__(self, filePath: str):
        self.__filePath = filePath

    @staticmethod
    @abc.abstractmethod
    def getAnalysisResultsClass() -> t_.Type[AbstractHDFAnalysisResults]:
        """

        Returns:
            The class that is used to contain analysis results for this acquisition type.
        """
        pass

    def getAnalyses(self) -> t_.List[str]:
        """

        Returns:
            A list of the names of analyses that were found.
        """
        assert self.__filePath is not None
        return self.getAnalysesAtPath(self.__filePath)

    @classmethod
    def getAnalysesAtPath(cls, path: str) -> t_.List[str]:
        """

        Args:
            path: The path to search for analysis files.

        Returns:
            A list of the names of analyses that were found.
        """
        anPath = os.path.join(path, 'analyses')
        if os.path.exists(anPath):
            files = os.listdir(os.path.join(path, 'analyses'))
            return [cls.getAnalysisResultsClass().fileName2Name(f) for f in files]
        else:
            return []

    def saveAnalysis(self, analysis: AbstractHDFAnalysisResults, name: str, overwrite: bool = False):
        """

        Args:
            analysis: An AnalysisResults object to be saved.
            name: The name to save the analysis as
            overwrite: If `True` then any existing file of the same name will be replaced. If `False` an exception will be raised.
        """
        path = os.path.join(self.__filePath, 'analyses')
        if not os.path.exists(path):
            os.mkdir(path)
        analysis.toHDF(path, name, overwrite=overwrite)

    def loadAnalysis(self, name: str) -> AbstractHDFAnalysisResults:
        """

        Args:
            name: The name of the analysis to load.

        Returns:
            A new instance of an AnalysisResults object.
        """
        return self.getAnalysisResultsClass().load(os.path.join(self.__filePath, 'analyses'), name)

    def removeAnalysis(self, name: str):
        """

        Args:
            name: The name of the analysis to be deleted
        """
        os.remove(os.path.join(self.__filePath, 'analyses', self.getAnalysisResultsClass().name2FileName(name)))


class DynMetaData(MetaDataBase, AnalysisManager):
    """A class that represents the metadata of a Dynamics acquisition."""
    class FileFormats(enum.Enum):
        """An enumerator identifying the types of file formats that this class can be loaded from."""
        Tiff = enum.auto()
        RawBinary = enum.auto()
        Hdf = enum.auto()

    @staticmethod
    def getAnalysisResultsClass() -> t_.Type[AbstractHDFAnalysisResults]:
        from pwspy.analysis.dynamics import DynamicsAnalysisResults
        return DynamicsAnalysisResults

    _jsonSchemaPath = os.path.join(_jsonSchemasPath, 'DynMetaData.json')
    with open(_jsonSchemaPath) as f:
        _jsonSchema = json.load(f)

    def __init__(self, metadata: dict, filePath: t_.Optional[str] = None, fileFormat: t_.Optional[FileFormats] = None, acquisitionDirectory: t_.Optional[Acquisition] = None):
        self.fileFormat = fileFormat
        MetaDataBase.__init__(self, metadata, filePath, acquisitionDirectory=acquisitionDirectory)
        AnalysisManager.__init__(self, filePath)

    def toDataClass(self, lock: mp.Lock = None) -> pwsdtd.DynCube:
        """
        Args:
            lock (mp.Lock): A multiprocessing `lock` that can prevent help us synchronize demands on the hard drive when loading many files in parallel. Probably not needed.
        Returns:
            pwsdtmd.DynCube: The data object associated with this metadata object.
        """
        # from pwspy.dataTypes.data import DynCube
        return pwsdtd.DynCube.fromMetadata(self, lock)

    @property
    def idTag(self) -> str:
        """

        Returns:
            str: A unique string identifying this acquisition.
        """
        return f"DynCube_{self.dict['system']}_{self.dict['time']}"

    @property
    def wavelength(self) -> int:
        """The wavelength that this acquisition was acquired at."""
        return self.dict['wavelength']

    @property
    def times(self) -> t_.Tuple[float, ...]:
        """A sequence indicatin the time associated with each 2D slice of the 3rd axis of the `data` array"""
        return self.dict['times']

    @classmethod
    def fromOldPWS(cls, directory: str, lock: mp.Lock = None, acquisitionDirectory: t_.Optional[Acquisition] = None) -> DynMetaData:
        """Loads old dynamics cubes which were saved the same as old pws cubes. a raw binary file with some metadata saved in random .mat files. Does not support
        automatic detection of binning, pixel size, camera dark counts, system name.

        Args:
            directory: The path to the folder containing the data files load the metadata from.

        Returns:
            A new instance of `DynMetaData`.
        """
        if lock is not None:
            lock.acquire()
        try:
            #While the info2 file exists for old dynamics acquisitions, it is just garbage char.
            info3 = list(spio.loadmat(os.path.join(directory, 'info3.mat'))['info3'].squeeze())
            wv = list(spio.loadmat(os.path.join(directory, 'WV.mat'))['WV'].squeeze())
            wv = [int(i) for i in wv]  # We will have issues saving later if these are numpy int types.
            assert all([i == wv[0] for i in wv]), "The wavelengths of the dynamics cube are not all identical."
            md = {
                #RequiredMetadata
                'exposure': info3[1],
                'time': '{:d}-{:d}-{:d} {:d}:{:d}:{:d}'.format(
                    *[int(i) for i in [info3[8], info3[7], info3[6], info3[9], info3[10], info3[11]]]),
                'system': str(info3[0]),
                'binning': None,
                'pixelSizeUm': None,
                'wavelength': wv[0],
                'times': [i*info3[1] for i in range(len(wv))],  # We don't have any record of the times so we just have to assume it matches exactly with the exposure time, this is in milliseconds.
                #Extra metadata
                'systemId': info3[0],
                'imgHeight': int(info3[2]), 'imgWidth': int(info3[3]), 'wavelengths': wv
                }
        finally:
            if lock is not None:
                lock.release()
        return cls(md, filePath=directory, fileFormat=DynMetaData.FileFormats.RawBinary, acquisitionDirectory=acquisitionDirectory)

    @classmethod
    def fromTiff(cls, directory, lock: mp.Lock = None, acquisitionDirectory: t_.Optional[Acquisition] = None) -> DynMetaData:
        """

        Args:
            directory: The path to the folder containing the data files load the metadata from.
        Returns:
            A new instance of `DynMetaData` loaded from file.
        """
        if lock is not None:
            lock.acquire()
        try:
            if os.path.exists(os.path.join(directory, 'dyn.tif')):
                path = os.path.join(directory, 'dyn.tif')
            else:
                raise OSError("No Tiff file was found at:", directory)
            if os.path.exists(os.path.join(directory, 'dynmetadata.json')):
                metadata = json.load(open(os.path.join(directory, 'dynmetadata.json'), 'r'))
            else:
                with tf.TiffFile(path) as tif:
                    metadata = json.loads(tif.imagej_metadata['Info'])  # The micromanager plugin saves metadata as the info property of the imagej imageplus object.
        finally:
            if lock is not None:
                lock.release()
        metadata['binning'] = metadata['MicroManagerMetadata']['Binning']['scalar']  # Get binning from the micromanager metadata
        metadata['pixelSizeUm'] = metadata['MicroManagerMetadata']['PixelSizeUm']['scalar']  # Get the pixel size from the micromanager metadata
        if metadata['pixelSizeUm'] == 0: metadata['pixelSizeUm'] = None
        return cls(metadata, filePath=directory, fileFormat=cls.FileFormats.Tiff, acquisitionDirectory=acquisitionDirectory)

    def getThumbnail(self) -> np.ndarray:
        """Return the image used for quick viewing of the acquisition. Has no numeric significance."""
        with tf.TiffFile(os.path.join(self.filePath, 'image_bd.tif')) as f:
            return f.asarray()


class ERMetaData:
    """A class representing the extra information related to an ExtraReflectanceCube file. This can be useful as a object
    to keep track of a ExtraReflectanceCube without having to have the data from the file loaded into memory.

    Args:
        inheritedMetadata: The metadata dictionary will often just be inherited information from one of the `PwsCubes` that was used to create
            this ER Cube. While this data can be useful it should be taken with a grain of salt. E.G. the metadata will contain
            an `exposure` field. In reality this ER Cube will have been created from pwsdtd.PwsCubes at a variety of exposures.
        numericalAperture: The numerical aperture that the PwsCubes used to generate this Extra reflection cube were imaged at.
        filePath: The path to the file that this object is stored in.
    """
    _jsonSchema = {"$schema": "http://json-schema.org/schema#",
                   '$id': 'extraReflectionMetadataSchema',
                   'title': 'extraReflectionMetadataSchema',
                   'required': ['system', 'time', 'wavelengths', 'pixelSizeUm', 'binning', 'numericalAperture'],
                   'type': 'object',
                   'properties': {
                       'system': {'type': 'string'},
                       'time': {'type': 'string'},
                       'wavelengths': {'type': 'array',
                                        'items': {'type': 'number'}
                                        },
                       'pixelSizeUm': {'type': ['number', 'null']},
                       'binning': {'type': ['integer', 'null']},
                       'numericalAperture': {'type': ['number']}
                        }
                   }
    FILESUFFIX = '_eReflectance.h5'
    DATASETTAG = 'extraReflection'
    _MDTAG = 'metadata'

    def __init__(self, inheritedMetadata: dict, numericalAperture: float, filePath: str = None):
        self.inheritedMetadata = inheritedMetadata
        self.inheritedMetadata['numericalAperture'] = numericalAperture
        jsonschema.validate(instance=inheritedMetadata, schema=self._jsonSchema, cls=_TupleValidator)
        self.filePath = filePath

    @property
    def idTag(self) -> str:
        """A unique tag to identify this acquisition by."""
        return f"ExtraReflection_{self.inheritedMetadata['system']}_{self.inheritedMetadata['time']}"

    @property
    def numericalAperture(self) -> float:
        """The numerical aperture that this cube was imaged at."""
        return self.inheritedMetadata['numericalAperture']

    @property
    def systemName(self) -> str:
        """The name of the system that this image was acquired on."""
        return self.inheritedMetadata['system']

    @classmethod
    def validPath(cls, path: str) -> t_.Tuple[bool, t_.Union[str, bytes], t_.Union[str, bytes]]:
        """

        Args:
            path: The file path to the file to search for valid ExtraReflectance files.

        Returns:
            A tuple containing: validPath: True if the path is valid, directory: The directory the file is in, name: The name that the object was saved as.
        """
        if cls.FILESUFFIX in path:
            directory, name = cls.directory2dirName(path)
            with h5py.File(os.path.join(directory, f'{name}{cls.FILESUFFIX}'), 'r') as hf:
                valid = cls._MDTAG in hf[cls.DATASETTAG].attrs
            return valid, directory, name
        else:
            return False, '', ''

    @classmethod
    def fromHdfFile(cls, directory: str, name: str) -> ERMetaData:
        """

        Args:
            directory: The directory the file is saved in.
            name: The name the object was saved as.

        Returns:
            A new instance of `ERMetaData` object
        """
        filePath = cls.dirName2Directory(directory, name)
        with h5py.File(filePath, 'r') as hf:
            dset = hf[cls.DATASETTAG]
            return cls.fromHdfDataset(dset, filePath=filePath)

    @classmethod
    def fromHdfDataset(cls, d: h5py.Dataset, filePath: str = None) -> ERMetaData:
        """

        Args:
            d: The `h5py.Dataset` to load the object from.

        Returns:
            A new instance of `ERMetaData` object
        """
        mdDict = json.loads(d.attrs[cls._MDTAG])
        return cls(mdDict, mdDict['numericalAperture'], filePath=filePath)

    def toHdfDataset(self, g: h5py.Group) -> h5py.Group:
        """

        Args:
            g: The `h5py.Group` to save the new dataset into.

        """
        g[self.DATASETTAG].attrs[self._MDTAG] = np.string_(json.dumps(self.inheritedMetadata))
        return g

    @classmethod
    def directory2dirName(cls, path: str) -> t_.Tuple[t_.Union[bytes, str], t_.Union[bytes, str]]:
        """

        Args:
            path: The path to the file that stores an `ExtraReflectanceCube` object.

        Returns:
            A tuple containing: directory: The directory path, name: The name that the file was saved as.
        """
        directory, fileName = os.path.split(path)
        if not fileName.endswith(cls.FILESUFFIX):
            raise ValueError(f"The file name \"{fileName}\" is not recognized as a {cls.__name__} file. Should end with \"{cls.FILESUFFIX}\".")
        name = fileName.split(cls.FILESUFFIX)[0]
        return directory, name

    @classmethod
    def dirName2Directory(cls, directory: str, name: str) -> str:
        """
        This is the inverse of `directory2dirName`
        """
        return os.path.join(directory, f'{name}{cls.FILESUFFIX}')


class FluorMetaData(MetaDataBase):
    """
    Metadata for a fluorescence image.

    Args:
        md: A dictionary containing the metadata

    """
    FILENAME = 'fluor.tif'
    MDPATH = 'fluorMetadata.json'

    _jsonSchemaPath = os.path.join(_jsonSchemasPath, 'MetaDataBase.json')
    with open(_jsonSchemaPath) as f:
        _jsonSchema = json.load(f)

    def __init__(self, md: dict, filePath: t_.Optional[str] = None, acquisitionDirectory: t_.Optional[Acquisition] = None):
        super().__init__(md, filePath, acquisitionDirectory)

    def toDataClass(self, lock: mp.Lock = None) -> pwsdtd.FluorescenceImage:
        return pwsdtd.FluorescenceImage.fromMetadata(self, lock)

    @property
    def idTag(self):
        return f"Fluor_{self.dict['system']}_{self.dict['time']}"

    @classmethod
    def fromTiff(cls, directory: str, acquisitionDirectory: t_.Optional[Acquisition]) -> FluorMetaData:
        """Load from a TIFF file.

        Args:
            directory: The path to the folder to load from.

        Returns:
            A new instance of `FluorMetaData` loaded from file.
        """
        if not FluorMetaData.isValidPath(directory):
            raise ValueError(f"Fluorescence image not found in {directory}.")
        with open(os.path.join(directory, FluorMetaData.MDPATH), 'r') as f:
            dic = json.load(f)

        # Get binning from the micromanager metadata
        binning = dic['MicroManagerMetadata']['Binning']['scalar']
        dic['binning'] = binning
        # Get the pixel size from the micromanager metadata
        try:
            dic['pixelSizeUm'] = dic['MicroManagerMetadata']['PixelSizeUm']['scalar']
        except KeyError:
            dic['pixelSizeUm'] = None
        if dic['pixelSizeUm'] == 0: dic['pixelSizeUm'] = None

        return FluorMetaData(dic, directory, acquisitionDirectory)

    @classmethod
    def isValidPath(cls, directory: str):
        """

        Args:
            directory: The path to search for valid files.

        Returns:
            True if a valid file was found.
        """
        path = os.path.join(directory, cls.FILENAME)
        path2 = os.path.join(directory, cls.MDPATH)
        return os.path.exists(path) and os.path.exists(path2)

    def getThumbnail(self) -> np.ndarray:
        """

        Returns:
            An image for quick viewing of the acquisition. No numerical significance.
        """
        with tf.TiffFile(os.path.join(self.filePath, 'image_bd.tif')) as f:
            return f.asarray()


class PwsMetaData(MetaDataBase, AnalysisManager):
    """A class that represents the metadata of a PWS acquisition.

    Args:
        metadata: The dictionary containing the metadata.
    """
    class FileFormats(enum.Enum):
        RawBinary = enum.auto()
        Tiff = enum.auto()
        Hdf = enum.auto()
        NanoMat = enum.auto()

    @staticmethod
    def getAnalysisResultsClass() -> t_.Type[AbstractHDFAnalysisResults]:
        from pwspy.analysis.pws import PWSAnalysisResults
        return PWSAnalysisResults

    _jsonSchemaPath = os.path.join(_jsonSchemasPath, 'PwsMetaData.json')
    with open(_jsonSchemaPath) as f:
        _jsonSchema = json.load(f)

    def __init__(self, metadata: dict, filePath: t_.Optional[str] = None, fileFormat: PwsMetaData.FileFormats = None, acquisitionDirectory: t_.Optional[Acquisition] = None):
        MetaDataBase.__init__(self, metadata, filePath, acquisitionDirectory=acquisitionDirectory)
        AnalysisManager.__init__(self, filePath)
        self.fileFormat: PwsMetaData.FileFormats = fileFormat
        self.dict['wavelengths'] = tuple(np.array(self.dict['wavelengths']).astype(float))

    def toDataClass(self, lock: mp.Lock = None) -> pwsdtd.PwsCube:
        return pwsdtd.PwsCube.fromMetadata(self, lock)

    @cached_property
    def idTag(self) -> str:
        return f"PwsCube_{self.dict['system']}_{self.dict['time']}"

    @property
    def wavelengths(self) -> t_.Tuple[float, ...]:
        return self.dict['wavelengths']

    @classmethod
    def loadAny(cls, directory, lock: mp.Lock = None, acquisitionDirectory: t_.Optional[Acquisition] = None) -> PwsMetaData:
        """
        Attempt to load from any file format.

        Args:
            directory: The file path to load the metadata from.
        Returns:
            A new instance of `PwsMetaData` loaded from file
        """
        try:
            return PwsMetaData.fromTiff(directory, lock=lock, acquisitionDirectory=acquisitionDirectory)
        except:
            try:
                return PwsMetaData.fromOldPWS(directory, lock=lock, acquisitionDirectory=acquisitionDirectory)
            except:
                try:
                    return PwsMetaData.fromNano(directory, lock=lock, acquisitionDirectory=acquisitionDirectory)
                except:
                    raise OSError(f"Could not find a valid PWS image cube file at {directory}.")

    @classmethod
    def fromOldPWS(cls, directory, lock: mp.Lock = None, acquisitionDirectory: t_.Optional[Acquisition] = None) -> PwsMetaData:
        """
        Attempt to load from the old .mat file format.

        Args:
            directory: The file path to load the metadata from.
        Returns:
            A new instance of `PwsMetaData` loaded from file
        """
        if lock is not None:
            lock.acquire()
        try:
            try:
                md = json.load(open(os.path.join(directory, 'pwsmetadata.txt')))
            except:  # have to use the old metadata
                info2 = list(spio.loadmat(os.path.join(directory, 'info2.mat'))['info2'].squeeze())
                info3 = list(spio.loadmat(os.path.join(directory, 'info3.mat'))['info3'].squeeze())
                logging.getLogger(__name__).info("Json metadata not found. Using backup metadata.")
                wv = list(spio.loadmat(os.path.join(directory, 'WV.mat'))['WV'].squeeze())
                wv = [int(i) for i in wv]  # We will have issues saving later if these are numpy int types.
                md = {'startWv': info2[0], 'stepWv': info2[1], 'stopWv': info2[2],
                      'exposure': info2[3], 'time': '{:d}-{:d}-{:d} {:d}:{:d}:{:d}'.format(
                        *[int(i) for i in [info3[8], info3[7], info3[6], info3[9], info3[10], info3[11]]]),
                      'systemId': info3[0], 'system': str(info3[0]),
                      'imgHeight': int(info3[2]), 'imgWidth': int(info3[3]), 'wavelengths': wv,
                      'binning': None, 'pixelSizeUm': None}
        finally:
            if lock is not None:
                lock.release()
        return cls(md, filePath=directory, fileFormat=PwsMetaData.FileFormats.RawBinary, acquisitionDirectory=acquisitionDirectory)

    @classmethod
    def fromNano(cls, directory: str, lock: mp.Lock = None, acquisitionDirectory: t_.Optional[Acquisition] = None) -> PwsMetaData:
        """
        Attempt to load from NanoCytomic .mat file format

        Args:
            directory: The file path to load the metadata from.
        Returns:
            A new instance of `PwsMetaData` loaded from file
        """
        if lock is not None:
            lock.acquire()
        try:
            with h5py.File(os.path.join(directory, 'imageCube.mat'), 'r') as hf:
                cubeParams = hf['cubeParameters']
                lam = cubeParams['lambda']
                exp = cubeParams['exposure'] #we don't support adaptive exposure.
                md = {'startWv': lam['start'][0, 0], 'stepWv': lam['step'][0, 0], 'stopWv': lam['stop'][0, 0],
                      'exposure': exp['base'][0, 0], 'time': datetime.strptime(np.string_(cubeParams['metadata']['date'][()].astype(np.uint8)).decode(), '%Y%m%dT%H%M%S').strftime(dateTimeFormat),
                      'system': np.string_(cubeParams['metadata']['hardware']['system']['id'][()].astype(np.uint8)).decode(), 'wavelengths': list(lam['sequence'][0]),
                      'binning': None, 'pixelSizeUm': None}
        finally:
            if lock is not None:
                lock.release()
        return cls(md, filePath=directory, fileFormat=PwsMetaData.FileFormats.NanoMat, acquisitionDirectory=acquisitionDirectory)

    @classmethod
    def fromTiff(cls, directory, lock: mp.Lock = None, acquisitionDirectory: t_.Optional[Acquisition] = None) -> PwsMetaData:
        """
        Attempt to load from the standard TIFF file format.

        Args:
            directory: The file path to load the metadata from.
        Returns:
            A new instance of `PwsMetaData` loaded from file
        """
        if lock is not None:
            lock.acquire()
        try:
            if os.path.exists(os.path.join(directory, 'MMStack.ome.tif')):
                path = os.path.join(directory, 'MMStack.ome.tif')
            elif os.path.exists(os.path.join(directory, 'pws.tif')):
                path = os.path.join(directory, 'pws.tif')
            else:
                raise OSError("No Tiff file was found at:", directory)
            if os.path.exists(os.path.join(directory, 'pwsmetadata.json')):
                metadata = json.load(open(os.path.join(directory, 'pwsmetadata.json'), 'r'))
            else:
                with tf.TiffFile(path) as tif:
                    try:
                        metadata = json.loads(tif.pages[0].description)
                    except:
                        metadata = json.loads(tif.imagej_metadata['Info'])  # The micromanager plugin saves metadata as the info property of the imagej imageplus object.
                    metadata['time'] = tif.pages[0].tags['DateTime'].value
        finally:
            if lock is not None:
                lock.release()
        if 'MicroManagerMetadata' not in metadata:  # Data saved by something other than the Micro-Manager acquisition plugin won't have this. I.E. NC  data saved by `toTiff`
            binning = metadata['binning']
            pixelSize = metadata['pixelSizeUm']
        else:
            #For a while the micromanager metadata was getting saved weird this fixes it.
            if 'major_version' in metadata['MicroManagerMetadata']:
                metadata['MicroManagerMetadata'] = metadata['MicroManagerMetadata']['map']
            # Get binning from the micromanager metadata
            binning = metadata['MicroManagerMetadata']['Binning']
            if isinstance(binning, dict):  # This is due to a property map change from beta to gamma
                binning = binning['scalar']
            # Get the pixel size from the micromanager metadata
            try:
                pixelSize = metadata['MicroManagerMetadata']['PixelSizeUm']['scalar']
            except KeyError:
                pixelSize = None
        metadata['binning'] = binning
        metadata['pixelSizeUm'] = pixelSize
        if metadata['pixelSizeUm'] == 0: metadata['pixelSizeUm'] = None
        if 'waveLengths' in metadata:  # Fix an old naming issue
            metadata['wavelengths'] = metadata['waveLengths']
            del metadata['waveLengths']
        return cls(metadata, filePath=directory, fileFormat=PwsMetaData.FileFormats.Tiff, acquisitionDirectory=acquisitionDirectory)

    def metadataToJson(self, directory):
        """
        Save the metadata to a JSON file.

        Args:
            directory: The folder path to save the new file to.

        """
        with open(os.path.join(directory, 'pwsmetadata.json'), 'w') as f:
            json.dump(self.dict, f)

    def getThumbnail(self) -> np.ndarray:
        """

        Returns:
            An image for quick viewing of the acquisition. No numerical significance.
        """
        if self.fileFormat == PwsMetaData.FileFormats.NanoMat:
            with h5py.File(os.path.join(self.filePath, 'image_bd.mat'), 'r') as hf:
                return np.array(hf['image_bd']).T.copy()  # For some reason these are saved transposed?
        else:
            with tf.TiffFile(os.path.join(self.filePath, 'image_bd.tif')) as f:
                return f.asarray()


class Acquisition:
    """This class handles the file structure of a single acquisition. this can include a PWS acquisition as well as colocalized Dynamics and fluorescence.

    Args:
        directory: the file path the root directory of the acquisition
    """
    def __init__(self, directory: t_.Union[str, os.PathLike]):
        self.filePath = os.path.abspath(directory)  # Forcing an absolute path helps by: A: normalizing the path so string comparisons work. B: making sure that if the object is pickled and then unpickled from a different working direcotory the path will still be valid if the data is unmoved.
        if (self.pws is None) and (self.dynamics is None) and (len(self.fluorescence) == 0):
            raise OSError(f"Could not find a valid PWS or Dynamics Acquisition at {directory}.")

    def __repr__(self):
        if len(self.filePath) > 20:
            path = ("%.15s" % self.filePath[::-1])[::-1]  # This is confusing. We cut everything but the last characters to make long paths readable.
            path = "..." + path  # Add ellipsis
        else:
            path = self.filePath
        return f"{self.__class__.__name__}({path})"

    @cached_property
    def pws(self) -> t_.Optional[PwsMetaData]:
        """PwsMetaData: Returns None if no PWS acquisition was found."""
        try:
            return PwsMetaData.loadAny(os.path.join(self.filePath, 'PWS'), acquisitionDirectory=self)
        except:
            try:
                return PwsMetaData.loadAny(os.path.join(self.filePath), acquisitionDirectory=self) #Many of the old files are saved here in the root directory.
            except:
                return None

    @cached_property
    def dynamics(self) -> t_.Optional[DynMetaData]:
        """DynMetaData: Returns None if no dynamics acquisition was found."""
        try:
            return DynMetaData.fromTiff(os.path.join(self.filePath, 'Dynamics'), acquisitionDirectory=self)
        except:
            try:
                return DynMetaData.fromOldPWS(self.filePath, acquisitionDirectory=self) #This is just for old acquisitions where they were saved in their own folder that was indistinguishable from a PWS acquisitison.
            except:
                return None

    @cached_property
    def fluorescence(self) -> t_.List[FluorMetaData]:
        """List[FluorMetaData]: Newer acquisitions allow for multiple fluorescence images saved to numbered subfolders"""
        i = 0
        imgs = []
        while True:
            path = os.path.join(self.filePath, f"Fluorescence_{i}")
            if not os.path.exists(path):
                break
            try:
                imgs.append(FluorMetaData.fromTiff(path, acquisitionDirectory=self))
            except ValueError:
                logging.getLogger(__name__).info(f"Failed to load fluorescence metadata at {path}")
            i += 1
        if len(imgs) == 0:  # No files were found.
            # Old files only had a single fluorescence image with no number on the folder name.
            path = os.path.join(self.filePath, 'Fluorescence')
            if os.path.exists(path):
                return [FluorMetaData.fromTiff(path, acquisitionDirectory=self)]
            else:
                return []
        else:
            return imgs

    @property
    def idTag(self):
        if self.pws is not None:
            return self.pws.idTag
        else: #We must have one of these two items.
            return self.dynamics.idTag

    def getRois(self) -> t_.List[t_.Tuple[str, int, RoiFile.FileFormats]]:
        """Return information about the Rois found in the acquisition's file path.
        See documentation for Roi.getValidRoisInPath()"""
        assert self.filePath is not None
        return RoiFile.getValidRoisInPath(self.filePath)

    def loadRoi(self, name: str, num: int, fformat: RoiFile.FileFormats = None) -> RoiFile:
        """Load a Roi that has been saved to file in the acquisition's file path."""

        assert isinstance(name, str), f"The ROI name must be a string. Got a {type(name)}"
        assert isinstance(num, int), f"The ROI number must be an integer. Got a {type(num)}"
        if fformat == RoiFile.FileFormats.MAT:
            return RoiFile.fromMat(self.filePath, name, num, acquisition=self)
        elif fformat == RoiFile.FileFormats.HDF3:
            return RoiFile.fromHDF(self.filePath, name, num, acquisition=self)
        elif fformat == RoiFile.FileFormats.HDF2:
            return RoiFile.fromHDF_legacy(self.filePath, name, num, acquisition=self)
        elif fformat == RoiFile.FileFormats.HDF:
            return RoiFile.fromHDF_legacy_legacy(self.filePath, name, num, acquisition=self)
        else:
            return RoiFile.loadAny(self.filePath, name, num, acquisition=self)

    def saveRoi(self, roiName: str, roiNumber: int, roi: Roi, overwrite: bool = False) -> RoiFile:
        """
        Save a Roi to file in the acquisition's file path.

        Args:
            roiName: The name to identify this ROI
            roiNumber: The number to identify this ROI
            roi: The ROI object defining the ROI geometry
            overwrite: If True then any existing ROIFile matching this name and number will be overwritten. Otherwise an OSError is raised.

        Returns:
            A reference to the new ROIFile

        Raises:
            OSError: If `overwrite` is False and an ROI of the same name and number
        already exists then an OSError will be raised.

        """
        return RoiFile.toHDF(roi, roiName, roiNumber, self.filePath, overwrite=overwrite, acquisition=self)

    def deleteRoi(self, name: str, num: int, fformat: t_.Optional[RoiFile.FileFormats] = None):
        RoiFile.deleteRoi(self.filePath, name, num, fformat=fformat)

    def editNotes(self):
        """Create a `notes.txt` file if it doesn't already exists and open it in a text editor."""
        filepath = os.path.join(self.filePath, 'notes.txt')
        filepath = os.path.normpath(filepath)  # Sometime there can be an error if we don't clean the file path like this.
        if not os.path.exists(filepath):
            with open(filepath, 'w') as f:
                pass
        if sys.platform.startswith('darwin'):
            subprocess.call(('open', filepath))
        elif os.name == 'nt':  # For Windows
            os.startfile(filepath)
        elif os.name == 'posix':  # For Linux, Mac, etc.
            subprocess.call(('xdg-open', filepath))

    def hasNotes(self) -> bool:
        """Indicates whether or not a `notes.txt` file was found."""
        return os.path.exists(os.path.join(self.filePath, 'notes.txt'))

    def getNotes(self) -> str:
        """Return the contents of `notes.txt` as a string."""
        if self.hasNotes():
            with open(os.path.join(self.filePath, 'notes.txt'), 'r') as f:
                return '\n'.join(f.readlines())
        else:
            return ''

    def getThumbnail(self) -> np.ndarray:
        """Return a thumbnail from any of the available acquisitions. Should be an 8bit normalized image."""
        if self.pws is not None:
            return self.pws.getThumbnail()
        elif self.dynamics is not None:
            return self.dynamics.getThumbnail()
        elif len(self.fluorescence) != 0:
            return self.fluorescence[0].getThumbnail()

    def getNumber(self) -> int:
        return int(self.filePath.split("Cell")[-1])

    def __setstate__(self, state):
        """When the object is unpickled this will check if the file path is still valid."""
        if not os.path.exists(state['filePath']):
            warnings.warn(f"{self.__class__.__name__} file path cannot be found. {state['filePath']}")
        self.__dict__ = state  # This is the default thing to do to unpicle the object.

    def __hash__(self) -> int:
        return hash(self.filePath)  # Since the class is intimiately tied to the filesystem this is all that makes it unique

    def __eq__(self, other: Acquisition) -> bool:
        if not isinstance(other, Acquisition):
            return False
        return self.filePath == other.filePath


_TupleValidator = jsonschema.validators.extend(  # All of this is just so that jsonschema will allow a tuple as a 'array' schema member.
    jsonschema.Draft7Validator,
    type_checker=jsonschema.Draft7Validator.TYPE_CHECKER.redefine(
        'array',
        lambda checker, instance: jsonschema.Draft7Validator.TYPE_CHECKER.is_type(instance, 'array') or isinstance(instance, tuple)
    ))

if __name__ == '__main__':
    md = PwsMetaData.fromNano(r'C:\Users\nicke\Desktop\LTL20b_Tracking cells in 50%EtOH,95%EtOH,Water\95% ethanol\Cell1')


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Functions for quickly loading files using parallel processing.

Functions
----------
.. autosummary::
   :toctree: generated/

   loadAndProcess
   processParallel

"""
__all__ = ['loadAndProcess', 'processParallel']

import logging
import multiprocessing as mp
import queue
import threading as th
from time import time
import typing
from typing import Union, Optional, List, Tuple
import pandas as pd
import psutil
from pwspy.dataTypes import Acquisition, MetaDataBase

'''Local Functions'''
def _load(loadHandle: Union[str, MetaDataBase], lock: mp.Lock):
    md: MetaDataBase
    if isinstance(loadHandle, str):
        md = Acquisition(loadHandle).pws # In the case that we just have a string to work with, we assume that we are loading a PWS file and not any other type such as dynamics.
    elif isinstance(loadHandle, MetaDataBase):
        md = loadHandle
    else:
        raise TypeError("files specified to the loader must be either str or inherited from pwspy.dataTypes.MetaDataBase")
    return md.toDataClass(lock)


def _loadIms(qout: queue.Queue, qin: queue.Queue, lock: th.Lock):
    """When not running in parallel this function is executed in a separate thread to load PwsCubes and populate a Queue
    with them."""
    logger = logging.getLogger(__name__)
    while not qin.empty():
        try:
            index, row = qin.get()
            displayStr = row['cube'].filePath if isinstance(row['cube'], MetaDataBase) else row['cube']
            logger.info(f'Starting {displayStr}')
            im = _load(row['cube'], lock=lock)
            row['cube'] = im
            qout.put((index, row), block=True)  # Once the queue is full we will block here so that we don't overfill the RAM.
            perc = psutil.virtual_memory().percent
            logger.info(f"Memory Usage: {perc}%")
            if perc >= 95:
                logger.info("Memory limit exceeded. Exiting to avoid lock-up.")
                return
        except Exception as e:
            qout.put(e)  # Put the error in the queue so it can propagate to the main thread.
            raise e


def _procWrap(procFunc, lock: th.Lock):
    def func(fromQueue, procFuncArgs=None):
        index, row = fromQueue
        im = row['cube']
        args = (im,)
        if procFuncArgs:
            ret = procFunc(*args, *procFuncArgs)
        else:
            ret = procFunc(*args)
        row['cube'] = ret
        return index, row
    return func


def _loadThenProcess(procFunc, procFuncArgs, lock: mp.Lock, row):
    """Handles loading the PwsCubes from file and if needed then calling the processorFunc. This function will be executed
     on each core when running in parallel. If not running in parallel then _loadIms will be used."""
    index, row = row
    im = _load(row['cube'], lock=lock)
    displayStr = row['cube'].filePath if isinstance(row['cube'], MetaDataBase) else row['cube']
    print("Run", displayStr, mp.current_process())
    ret = procFunc(im, *procFuncArgs)
    row['cube'] = ret
    return row


'''User Functions'''


def loadAndProcess(fileFrame: Union[pd.DataFrame, List, Tuple], processorFunc: Optional = None, parallel: Optional = None,
                   procArgs: Optional = None, initializer=None,
                   initArgs=None) -> Union[pd.DataFrame, List, Tuple]:
    """DEPRECATED! This over-complicated function should be replaced with usage of processParallel.
    A convenient function to load a series of Data Cubes from a list or dictionary of file paths.

    Parameters
    ----------
    fileFrame
        A dataframe containing a column of PwsCube file paths titled 'cube' and other columns to act as specifiers for each cube.
        If no specifiers are used this can just be a list of file paths.
    processorFunc
        A function that each loaded cell should be passed to. The first argument of processorFunc should be the loaded
        PwsCube. Additional arguments can be passed to processorFunc using the procArgs variable.
    parallel
        default is `False`. If `True` then the loading and processing will be performed in parallel on multiple cores,
        otherwise it will be done using multithreading on a single core. Setting this to true can result if big speedups
        if the time to run processorFunc is greater than the time to load an PwsCube from file.
    procArgs
        Optional arguments to pass to processorFunc
    initializer:
        A function that is run once at the beginning of each spawned process. Can be used for copying shared memory.
    initArgs:
        A tuple of arguments to pass to the `initializer` function.

    Returns
    -------
        An object of the same form as fileFrame except the PwsCube file paths have been replaced by PwsCube Object.
        If using processorFunc the return values from processorFunc will be returned rather than PwsCube Objects.
    """
    if procArgs is None:
        procArgs = []
    if parallel is None:
        parallel = False if processorFunc is None else True  # No reason to run in parallel if we don't have a computationally expensive function to run.
    #If the fileFrame provided is not already a pandas dataframe the convert and store a reference to the original type. We'll try to convert back at the end.
    origClass = None
    if not isinstance(fileFrame, pd.DataFrame):
        try:
            origClass = type(fileFrame)
            fileFrame = pd.DataFrame({'cube': fileFrame})
        except:
            raise TypeError("fileFrame cannot be converted to a pandas dataframe")
    # Error checking
    if 'cube' not in fileFrame.columns:
        raise IndexError("The fileFrame must contain a 'cube' column.")
    logging.getLogger(__name__).info(f"Starting loading {len(fileFrame)} files.")
    sTime = time()
    if parallel:
        if processorFunc is None:
            raise Exception("Running in parallel with no processorFunc is pointles. Set parallel to False to run in multithreaded mode.")
        m = mp.Manager()
        lock = m.Lock()
        numProcesses = psutil.cpu_count(logical=False) - 1  # Use one less than number of available cores.
        po = mp.Pool(processes=numProcesses, initializer=initializer, initargs=initArgs)
        try:
            cubes = po.starmap(_loadThenProcess, zip(*zip(*[[processorFunc, procArgs, lock]] * len(fileFrame)), fileFrame.iterrows()))
        finally:
            po.close()
            po.join()
    else:
        if initializer:
            initializer(*initArgs)
        qout = queue.Queue(maxsize=3)  # Once 3 cells are loaded into the queue it will block so that they can be processed. Prevents us from using way to much RAM.
        qin = queue.Queue()
        [qin.put(f) for f in fileFrame.iterrows()]
        lock = th.Lock()
        thread = th.Thread(target=_loadIms, args=[qout, qin, lock])
        thread.start()
        cubes = []
        if processorFunc:
            wrappedFunc = _procWrap(processorFunc, lock)
        for i in range(len(fileFrame)):
            ret = qout.get()
            if isinstance(ret, Exception):
                raise ret
            else:
                if processorFunc:
                    cubes.append(wrappedFunc(ret, procArgs))
                else:
                    cubes.append(ret)  # A list of tuples of index, dataframe row
        thread.join()
        indices, cubes = zip(*sorted(cubes)) #This ensures that the return value is in the same order as the input array.
    logging.getLogger(__name__).info(f"Loading took {time() - sTime} seconds")
    ret = pd.DataFrame(list(cubes))
    if origClass is None:
        return ret
    else:
        return origClass(ret['cube'])


def processParallel(fileFrame: pd.DataFrame, processorFunc: typing.Callable[[], typing.Any], initializer: typing.Callable=None, initArgs: Tuple=None, procArgs: Tuple=None, numProcesses: int = None) -> List:
    """A convenience function to process the rows of a pandas DataFrame in parallel

    Parameters
    ----------
    fileFrame
        A dataframe. Each row of the frame will be passed as the first argument to the processorFunc.
    processorFunc
        A function that each row number and row of the `fileFrame` should be passed to as the first and second argument respectively. Additional arguments can
        be passed to processorFunc using the procArgs variable. The function should return the value which you want included
        in the return of `processParrallel`.
    procArgs
        Optional arguments to pass to processorFunc
    initializer:
        A function that is run once at the beginning of each spawned process. Can be used for copying shared memory.
    initArgs:
        A tuple of arguments to pass to the `initializer` function.

    Returns
    -------
        List containing the results of each execution of `processorFunc`.
    """
    if numProcesses is None:
        numProcesses = psutil.cpu_count(logical=False) - 1  # Use one less than number of available cores. If we use all cores then things can get locked up.
    po = mp.Pool(processes=numProcesses, initializer=initializer, initargs=initArgs)
    try:
        vars = fileFrame.iterrows() if procArgs is None else zip(*zip(*fileFrame.iterrows()), *zip(*[procArgs] * len(fileFrame)))
        cubes = po.starmap(processorFunc, vars)
    finally:
        po.close()
        po.join()
    return cubes


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

import tifffile as tf
from glob import glob
import os


class Image:
    """Represents a multi-file Tiff image saved by Micro-Manager

    Args:
        directory: The file path to the folder containing the Micro-Manager TIFF files.
    """
    def __init__(self, directory: str):
        files = glob(os.path.join(directory, '*MMStack*'))
        self.f = tf.TiffFile(files[0])

    def getPosition(self, x: int, y: int):
        return [im for im in self.f.series if tuple([int(i) for i in im.name.split('-Pos')[-1].split('_')]) == (y, x)][0]

if __name__ == '__main__':
    a = Image(r'G:\Data\waterdishcoverage\medconf\fixed_1')


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
Classes used during the "compilation" step of analysis. This is when the data saved during analysis
is combined with ROIs to generate a table of values such as the average RMS, reflectance, diffusion coefficient, etc.

PWS
----------
.. autosummary::
   :toctree: generated/

   PWSCompilerSettings
   PWSRoiCompilationResults
   PWSRoiCompiler

Dynamics
------------
.. autosummary::
    :toctree: generated/

    DynamicsCompilerSettings
    DynamicsRoiCompilationResults
    DynamicsRoiCompiler

Generic
--------
.. autosummary::
    :toctree: generated/

    GenericCompilerSettings
    GenericRoiCompilationResults
    GenericRoiCompiler

"""

__all__ = ['DynamicsRoiCompiler', 'DynamicsRoiCompilationResults', 'DynamicsCompilerSettings',
           'PWSRoiCompiler', 'PWSRoiCompilationResults', 'PWSCompilerSettings', 'GenericRoiCompiler',
           'GenericRoiCompilationResults', 'GenericCompilerSettings', 'AbstractRoiCompilationResults',
           'AbstractRoiCompiler', 'AbstractCompilerSettings']

from ._dynamics import DynamicsCompilerSettings, DynamicsRoiCompilationResults, DynamicsRoiCompiler
from ._pws import PWSCompilerSettings, PWSRoiCompilationResults, PWSRoiCompiler
from ._generic import GenericCompilerSettings, GenericRoiCompilationResults, GenericRoiCompiler
from ._abstract import AbstractCompilerSettings, AbstractRoiCompilationResults, AbstractRoiCompiler



import pathlib as pl
import pwspy.dataTypes as pwsdt
import pytest


testDataPath = pl.Path(__file__).parent / 'resources' / 'test_data'  # The path to find the test data in.


class Dataset:
    """
    Just a simple class to represent what is needed from a single test dataset.
    """
    datasetPath: pl.Path   # The path to the folder containing all the acquisitions
    referenceCellPath: pl.Path  # The path to the acquisition folder that is to be used as the reference for analysis.

    def __init__(self, datasetPath: pl.Path, referenceCellPath: pl.Path):
        self.datasetPath = datasetPath
        self.referenceCellPath = referenceCellPath

    def clean(self):
        """
        When starting the tests we don't need any analysis files which take a lot of space.
        go through and delete all analysis files.
        """
        acqs = [pwsdt.Acquisition(i) for i in self.datasetPath.glob("Cell[0-9]")]
        for acq in acqs:
            if acq.pws:
                for anName in acq.pws.getAnalyses():
                    acq.pws.removeAnalysis(anName)
                    print(f"Deleted {acq.filePath} PWS {anName}")
            if acq.dynamics:
                for anName in acq.dynamics.getAnalyses():
                    acq.dynamics.removeAnalysis(anName)
                    print(f"Deleted {acq.filePath} Dynamics {anName}")


@pytest.fixture(scope='session')
def sequenceData() -> Dataset:
    """A reference to the dataset with PWS data collected with the automated event sequencing plugin. Cleans out analysis files when done."""
    ds = Dataset(
        testDataPath / 'sequencer',
        testDataPath / 'sequencer' / "Cell3"
    )
    yield ds
    ds.clean()


@pytest.fixture(scope='session')
def dynamicsData() -> Dataset:
    """A reference to the dataset with PWS and Dynamics data. Cleans out analysis files when done."""
    ds = Dataset(
        testDataPath / 'dynamics',
        testDataPath / 'dynamics' / "Cell999"
    )
    yield ds
    ds.clean()



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Provides a number of functions for calculating simple reflections based on known refractive indices

Functions
----------
.. autosummary::
   :toctree: generated/

   getReflectance
   getRefractiveIndex

"""

__all__ = ['getReflectance', 'getRefractiveIndex']

import typing
from numbers import Number

import pandas as pd
import numpy as np
import os
from pwspy.utility.reflection import Material

from pwspy.utility.reflection.multilayerReflectanceEngine import Stack, Layer, Polarization
from typing import Union, List, Tuple

# File paths for materials that have the N values stored in csv.
materialFiles = {
    Material.Glass: 'N-BK7.csv',
    Material.Water: 'Daimon-21.5C.csv',
    Material.Air: 'Ciddor.csv',
    Material.Silicon: 'Silicon.csv',
    Material.Oil_1_7: 'CargilleOil1_7.csv',
    Material.Oil_1_4: "CargilleOil1_4.csv",
    Material.Ipa: 'Sani-DellOro-IPA.csv',
    Material.Ethanol: 'Rheims.csv',
    Material.ITO: 'Konig.csv',
    Material.Methanol: 'Moutzouris.csv'
}

# Coefficients for materials that use the Cauchy equation
materialCoeffs = {
    Material.Glycerol_99_5: (1.45797, 0.00598, -0.00036)  # 1.45797+0.00598λ−2−0.00036λ−4. Rheims et al. 1997: n 0.589-1.05 µm
}


def _init():
    """This function initializes data from files when the module is imported."""
    fileLocation = os.path.join(os.path.split(__file__)[0], 'refractiveIndexFiles')
    ser = {}  # a dictionary of the series by name
    for name, file in materialFiles.items():
        # create a series for each csv file
        arr = np.genfromtxt(os.path.join(fileLocation, file), skip_header=1, delimiter=',')
        _ = pd.DataFrame({'n': arr[:, 1], 'k': arr[:, 2]}, index=arr[:, 0].astype(np.float) * 1e3)
        ser[name] = _

    # Find the first and last indices that won't require us to do any extrapolation
    first = []
    last = []
    for k, v in ser.items():
        first += [v.first_valid_index()]
        last += [v.last_valid_index()]
    first = max(first)  # The lowest wavelength we have data for all materials.
    last = min(last)  # The higherst wavelength we have data for all materials.
    # Interpolate so we don't have any nan values.
    #    df = pd.DataFrame(ser)
    df = pd.concat(ser, axis='columns', keys=materialFiles.keys())
    df = df.interpolate('index')
    df = df.loc[first:last]  # Crop to the first and last wavelengths that we have data for all materials.

    # Calculate from the cauchy Equation
    wavelengths = df.index / 1e3  # Convert to microns
    for name, coeffs in materialCoeffs.items():
        arr = np.zeros((len(wavelengths),))
        for i, coeff in enumerate(coeffs):
            arr += coeff / (wavelengths ** (i*2))
        df[(name, 'n')] = arr
        df[(name, 'k')] = 0

    return df

_n = _init()
del _init


def getRefractiveIndex(mat: Material, wavelengths: typing.Optional[typing.Iterable[float]] =None) -> pd.Series:
    """Get the spectrally dependent refractive index of a material.

    Args:
        mat: The material the retrieve the refractive index of.
        wavelengths: The wavelengths that the refractive index should be calculated at. If left as `None` then the wavelengths
            used will be determined by the original file that the data was pulled from.

    Returns:
        The refractive index. The index of the pandas series is the wavelengths.
    """
    refractiveIndex = np.array([np.complex(i[0], i[1]) for idx, i in _n[mat].iterrows()])
    refractiveIndex = pd.Series(refractiveIndex, _n.index)
    if wavelengths is not None: #Need to do interpolation
        wavelengths = pd.Index(wavelengths)
        combinedIdx = refractiveIndex.index.append(
            wavelengths)  # An index that contains all the original index points and all of the new. That way we can interpolate without first throwing away old data.
        from scipy.interpolate import griddata
        out = griddata(refractiveIndex.index, refractiveIndex.values, wavelengths)  #This works with complex numbers
        refractiveIndex = pd.Series(out, index=wavelengths)
    return refractiveIndex


def getReflectance(mat1: Union[Number, pd.Series, Material], mat2: Union[Number, pd.Series, Material], wavelengths: Union[np.ndarray, List, Tuple] = None, NA: float = 0) -> pd.Series:
    """Given the names of two interfaces this provides the reflectance in units of percent.
    If given a series as wavelengths the data will be interpolated and reindexed to match the wavelengths.

    Args:
        mat1: The first material comprising the reflective interface. This can either be a number or series of numbers representing the refractive index at different wavelengths
            or it can be a `Material` in which case the refractive index will be automatically calculated.
        mat2: The second material comprising the reflective interface. This can either be a number or series of numbers representing the refractive index at different wavelengths
            or it can be a `Material` in which case the refractive index will be automatically calculated.
        wavelengths: The wavelengths to calculate the reflectance at.
        NA: The numerical aperture of the system. Reflectance will be calculated by radially integrating results over
            the range of angles present within the numerical aperture. If left as `None` the result is calculated for
            light with a 0 degree angle of incidence.

    Returns:
        The percentage reflectance. The index of the pandas Series is the wavelengths.
    """
    index = _n.index if wavelengths is None else wavelengths
    if isinstance(index, Number):
        index = np.array([index])
    elif not isinstance(index, np.ndarray):
        index = np.array(index)
    s = Stack(wavelengths=index)
    s.addLayer(Layer(mat1, 1e9))  # Add a meter thick layer
    s.addLayer(Layer(mat2, 1e9))
    if NA == 0:
        d = s.calculateReflectance(np.array([0]))
        r = (d[Polarization.TE] + d[Polarization.TM]) / 2
        r = pd.Series(r.squeeze(), index=index)  # Get reflectance at 0 NA (0 incident angle)
    else:
        r = s.circularIntegration(np.linspace(0, NA, num=1000))
    return r

if __name__ == '__main__':
    a = 1


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
A collection of functions dedicated to the purpose of generating Extra Reflectance calibrations from images of
materials with known reflectances (e.g. air/glass interface, water/glass interface.)

By calculating the "extra reflectance" of a microscope system we can come up with a subtraction from our raw data that
will make our ratiometric measurements proportional to the actual sample reflectance.

These functions are relied on heavily in "ERCreator" app found in `pwspy_gui.ExtraReflectanceCreator`.

Functions
----------
.. autosummary::
   :toctree: generated/

   getTheoreticalReflectances
   generateMaterialCombos
   getAllCubeCombos
   plotExtraReflection
   generateRExtraCubes

Classes
--------
.. autosummary
   :toctree: generated/

   CubeCombo
"""
import logging
import typing as t_

import pwspy.dataTypes as pwsdt
from pwspy.utility.reflection import reflectanceHelper, Material
import itertools
import matplotlib.pyplot as plt
import numpy as np
from functools import reduce
import pandas as pd
from dataclasses import dataclass
import matplotlib as mpl
__all__ = ['getAllCubeCombos', 'generateMaterialCombos', 'generateRExtraCubes', 'getTheoreticalReflectances']

MCombo = t_.Tuple[Material, Material]  # This is just an alias to shorten some of the type hinting.


class CubeCombo:
    """A convenient way of packaging together two PwsCubes along with the material that each PwsCube is an image of.

    Args:
        material1: The material that was imaged in `cube1`
        material2: The material that was imaged in `cube2`
        cube1: An PwsCube with an image of a glass-{material} interface where the material is `material1`.
        cube2: An PwsCube with an image of a glass-{material} interface where the material is `material2`.

    """
    def __init__(self, material1: Material, material2: Material, cube1: pwsdt.PwsCube, cube2: pwsdt.PwsCube):
        self.mat1 = material1
        self.mat2 = material2
        self.data1 = cube1
        self.data2 = cube2

    def keys(self) -> MCombo:
        return self.mat1, self.mat2

    def values(self) -> t_.Tuple[pwsdt.PwsCube, pwsdt.PwsCube]:
        return self.data1, self.data2

    def items(self) -> t_.Iterator[t_.Tuple[Material, pwsdt.PwsCube]]:
        return zip(self.keys(), self.values())

    def __getitem__(self, item: Material) -> pwsdt.PwsCube:
        """cubeCombo[myMaterial] will return the data class for that material."""
        if item == self.mat1:
            return self.data1
        elif item == self.mat2:
            return self.data2
        else:
            raise KeyError("Key must be either mat1 or mat2.")


@dataclass
class _ComboSummary:
    """A convenient packaging of information related to the extra reflectance calculated from a single _CubeCombo"""
    mat1Spectra: np.ndarray  # The average spectrum in the ROI
    mat2Spectra: np.ndarray
    rExtra: np.ndarray
    I0: np.ndarray
    weight: np.ndarray


def getTheoreticalReflectances(materials: t_.Set[Material], wavelengths: t_.Tuple[float], numericalAperture: float) -> t_.Dict[Material, pd.Series]:
    """Generate a dictionary containing a Pandas `Series` of the `material`-glass reflectance for each material in
    `materials`.

    Args:
        materials: The set of materials that you want to retrieve the theoretical reflectance for.
        wavelengths: The wavelengths that you want the reflectances calculated at.
        numericalAperture: The numerical aperture that the reflectance should be calculated at.

    Returns:
        A dictionary of the reflectances for each material. The material serves as the dictionary key.
    """
    theoryR = {}
    for material in materials:  # For each unique material
        logging.getLogger(__name__).info(f"Calculating reflectance for {material}")
        theoryR[material] = reflectanceHelper.getReflectance(material, Material.Glass, wavelengths=wavelengths, NA=numericalAperture)
    return theoryR


def generateMaterialCombos(materials: t_.Iterable[Material], excludedCombos: t_.Optional[t_.Iterable[MCombo]] = None) -> t_.List[MCombo]:
    """Given a list of materials, this function returns a list of all possible material combo tuples.

    Args:
        materials: The list of materials that you want to generate every possible combo of.
        excludedCombos: Combinations of materials that you don't want included in the combinations.

    Returns:
        A list of Material combinations.
    """
    if excludedCombos is None:
        excludedCombos = []
    matCombos = list(itertools.combinations(materials, 2))  # All the 2-combinations of materials that can be compared
    matCombos = [(m1, m2) for m1, m2 in matCombos if
                 not (((m1, m2) in excludedCombos) or ((m2, m1) in excludedCombos))]  # Filter out all excluded combinations.
    for i, (m1, m2) in enumerate(matCombos):  # Make sure to arrange materials so that our reflectance ratio is greater than 1
        if (reflectanceHelper.getReflectance(m1, Material.Glass) / reflectanceHelper.getReflectance(m2, Material.Glass)).mean() < 1:
            matCombos[i] = (m2, m1)  # Swap elements if the ratio of reflectance is <1
    return matCombos


def getAllCubeCombos(matCombos: t_.Iterable[MCombo], cubeDict: t_.Dict[Material, t_.List[pwsdt.PwsCube]]) -> t_.Dict[MCombo, t_.List[CubeCombo]]:
    """Given a list of material combo tuples, return a dictionary whose keys are the material combo tuples and whose values are
    lists of CubeCombos.

    Args:
        matCombos: A list of material combinations, most likely generated by `generateMaterialCombos`
        cubeDict: An dictionary containing lists of a PwsCube measuremnts keyed by the material they were measured at.

    Returns:
        A dictionary with a key for each material combination. Each value is a list of all the `CubeCombo`s extracted from `cubes`.
    """
    allCombos = {}  # A dictionary where each key is a material combo, e.g. (mat1, mat2). The values are lists of every possible 2-combination of PwsCube measuremens for that material combo.
    for matCombo in matCombos:
        # Very important that The CubeCombo has data in the same order that the matCombo is in. Otherwise our data is all mismatched.
        matCubes = (
            cubeDict[matCombo[0]],
            cubeDict[matCombo[1]]
        )
        allCombos[matCombo] = [CubeCombo(*matCombo, *combo) for combo in itertools.product(*matCubes)]
    allCombos = {key: val for key, val in allCombos.items() if len(val) > 0}  # In some cases a matCombo appears for which there is no data. get rid of these.
    return allCombos


def _calculateSpectraFromCombos(cubeCombos: t_.Dict[MCombo, t_.List[CubeCombo]], theoryR: t_.Dict[Material, pd.Series],
                                 mask: t_.Optional[pwsdt.Roi] = None) ->\
        t_.Tuple[
            _ComboSummary,
            t_.Dict[MCombo, _ComboSummary],
            t_.Dict[MCombo, t_.List[t_.Tuple[_ComboSummary, CubeCombo]]]
        ]:
    """This is used to examine the output of extra reflection calculation before using saveRExtra to save a cube for each setting.
    Expects a dictionary as created by `getAllCubeCombos` and a dictionary of theoretical reflections.

    Args:
        cubeCombos: A dictionary containing all possible combinations of PwsCubes measured at different materials. Keyed by the material combo.
        theoryR: A dictionary containing the theoretical reflectances for all `Materials` in use. Should be accurate for the `numericalAperture` in question.
        mask: An ROI that limits the region of the PwsCubes that is analyzed. The spectra will be averaged over this region. If `None` the spectra will be average over the full XY FOV.

    Returns:
        The first item is a dictionary containing information about the average calculation for each material combination
        as well as the average calculation accross all material combinations. The seconds item is a dictionary containing
        information about every single cube combo.
    """
    # Generate summaries for every possible combination of measurements.
    allComboSummary: t_.Dict[MCombo, t_.List[t_.Tuple[_ComboSummary, CubeCombo]]] = {}  # Organize combos by the material combo they go with.
    for matCombo in cubeCombos.keys():
        allComboSummary[matCombo] = []
        for combo in cubeCombos[matCombo]:
            mat1, mat2 = combo.keys()
            spectra1 = combo[mat1].getMeanSpectra(mask)[0]
            spectra2 = combo[mat2].getMeanSpectra(mask)[0]
            weight = (spectra1 - spectra2) ** 2 / (spectra1 ** 2 + spectra2 ** 2) # See `_generateOneRExtraCube` for an explanation of this weighting.
            rExtra = ((theoryR[mat1] * spectra2) - (theoryR[mat2] * spectra1)) / (spectra1 - spectra2)
            I0 = spectra2 / (theoryR[mat2] + rExtra)  # Reconstructed intensity of illumination in same units as `spectra`. This could just as easily be done with material1. They are identical by definition.
            c = _ComboSummary(mat1Spectra=spectra1,
                              mat2Spectra=spectra2,
                              weight=weight,
                              rExtra=rExtra,
                              I0=I0)
            allComboSummary[matCombo].append((c, combo))

    # Calculate the averages for each material combo.
    meanComboSummary: t_.Dict[MCombo, t_.Any] = {}
    params = ('rExtra', 'I0', 'mat1Spectra', 'mat2Spectra')  # attribute names of _CubeCombo that are 1d arrays. Used for looping through and generating averages.
    for matCombo, comboSummaries in allComboSummary.items():
        weights = np.array([comboSummary.weight for comboSummary, _ in comboSummaries])
        spectra1 = np.average(np.array([comboSummary.mat1Spectra for comboSummary, _ in comboSummaries]), axis=0, weights=weights)
        spectra2 = np.average(np.array([comboSummary.mat2Spectra for comboSummary, _ in comboSummaries]), axis=0, weights=weights)
        I0 = np.average(np.array([comboSummary.I0 for comboSummary, _ in comboSummaries]), axis=0, weights=weights)
        rExtra = np.average(np.array([comboSummary.rExtra for comboSummary, _ in comboSummaries]), axis=0, weights=weights)
        meanSummary = _ComboSummary(
            mat1Spectra=spectra1,
            mat2Spectra=spectra2,
            weight=np.mean(weights, axis=0),
            rExtra=rExtra,
            I0=I0
        )
        meanComboSummary[matCombo] = meanSummary

    # Calculate the average accross all material combos.
    weights = np.array([comboSummary.weight for comboSummary in meanComboSummary.values()])
    totalMean = _ComboSummary(
        mat1Spectra=np.average(np.array([comboSummary.mat1Spectra for comboSummary in meanComboSummary.values()]), axis=0, weights=weights),
        mat2Spectra=np.average(np.array([comboSummary.mat2Spectra for comboSummary in meanComboSummary.values()]), axis=0, weights=weights),
        weight=np.mean(weights, axis=0),
        rExtra=np.average(np.array([comboSummary.rExtra for comboSummary in meanComboSummary.values()]), axis=0, weights=weights),
        I0=np.average(np.array([comboSummary.I0 for comboSummary in meanComboSummary.values()]), axis=0, weights=weights)
    )
    return totalMean, meanComboSummary, allComboSummary


def plotExtraReflection(images: t_.Dict[str, t_.Dict[Material, t_.List[pwsdt.PwsCube]]], theoryR: t_.Dict[Material, pd.Series], matCombos: t_.List[MCombo],
                        mask: t_.Optional[pwsdt.Roi] = None) -> t_.List[plt.Figure]:
    """Generate a variety of plots displaying information about the extra reflectance calculation.
    
    Args:
        images: A dictionary where the keys are strings representing some configuration of the system and the
            values are dictionaries where the keys are a `Material` and the values are lists of the `PwsCube` that were
            measured at the corresponding glass-{material} interface and configuration indicated by the dictionary keys.
        theoryR: A dictionary where the key is a `Material` and the value is a Pandas 'Series' giving the reflectance for
            a glass-{material} reflection over a range of wavelengths. The index of the series should be the wavelengths.
        matCombos: A list of the various material combinations that should be evaluated.
        mask: An ROI indicating the region of the images that should be included in the evaluation.

    Returns:
        A list of matplotlib figures resulting from this calculation.

    """
    from mpl_qt_viz.visualizers import DockablePlotWindow  # This is not a dependency of the library, just needed for this method.

    settings = set(images.keys())
    totalMean: t_.Dict[str, _ComboSummary] = {}
    meanValues: t_.Dict[str, t_.Dict[MCombo, _ComboSummary]] = {}
    allCombos: t_.Dict[str, t_.Dict[MCombo, t_.List[t_.Tuple[_ComboSummary, CubeCombo]]]] = {}
    for setting, materialCubeDict in images.items():
        cubeCombos = getAllCubeCombos(matCombos, materialCubeDict)
        totalMean[setting], meanValues[setting], allCombos[setting] = _calculateSpectraFromCombos(cubeCombos, theoryR, mask)

    dock = DockablePlotWindow("Primary")
    figs = [dock]
    fig, ax = dock.subplots("System Reflectance")  # For extra reflections
    ax.set_ylabel("System Reflectance % (100 = total reflection)")
    ax.set_xlabel("Wavelength (nm)")
    numLines = []
    for sett in settings:
        for matCombo in allCombos[sett].keys():
            numLines.append(len(allCombos[sett][matCombo]))
    numLines = sum(numLines)
    colormap = plt.cm.gist_rainbow
    colorCycle = mpl.cycler(color=plt.cm.gist_rainbow(np.linspace(0, 0.99, numLines)))
    ax.set_prop_cycle(colorCycle)
    for sett in settings:
        for matCombo in allCombos[sett].keys():
            mat1, mat2 = matCombo
            for comboSummary, combo in allCombos[sett][matCombo]:
                cubes = combo
                ax.plot(cubes[mat1].wavelengths, comboSummary.rExtra * 100,
                        label=f'{sett} {mat1.name} : {mat2.name}')
        ax.plot(cubes[mat1].wavelengths, totalMean[sett].rExtra * 100, color='k', label=f'{sett} Weighted Avg.')  # TODO Add a hover annotation since all of the lines are black it's impossible to know which one is which.
    ax.legend()

    fig2, ratioAxes = dock.subplots("Reflectance Ratios",
                                    subplots_kwargs=dict(nrows=len(matCombos))
                                    )  # for correction factor
    if not isinstance(ratioAxes, np.ndarray):
        ratioAxes = np.array(ratioAxes).reshape(1)  # If there is only one axis we still want it to be a list for the rest of the code
    ratioAxes = dict(zip(matCombos, ratioAxes))
    for combo in matCombos:
        ratioAxes[combo].set_title(f'{combo[0].name}/{combo[1].name} reflection ratio')
        ratioAxes[combo].plot(theoryR[combo[0]] / theoryR[combo[1]], label='Theory')
    for sett in settings:
        for matCombo in allCombos[sett].keys():
            mat1, mat2 = matCombo
            for comboSummary, combo in allCombos[sett][matCombo]:
                cubes = combo
                ratioAxes[matCombo].plot(cubes[mat1].wavelengths, comboSummary.mat1Spectra / comboSummary.mat2Spectra,
                                         label=f'{sett} {mat1.name}:{int(cubes[mat1].metadata.exposure)}ms {mat2.name}:{int(cubes[mat2].metadata.exposure)}ms')
    [ratioAxes[combo].legend() for combo in matCombos]

    comparisonDock = DockablePlotWindow("Correction Comparison")
    figs.append(comparisonDock)
    for sett in settings:
        settMatCombos = allCombos[sett].keys()  # Sometime we are looking at settings which don't have all the same matCombos. Only use the combos specific to this setting.
        means = totalMean[sett]
        fig6, scatterAx3 = comparisonDock.subplots(f"{sett} Uncorrected")
        scatterAx3.set_ylabel("Theoretical Ratio")
        scatterAx3.set_xlabel("Observed Ratio. No correction")
        scatterPointsY = [(theoryR[matCombo[0]] / theoryR[matCombo[1]]).mean() for matCombo in settMatCombos]
        scatterPointsX = [(meanValues[sett][matCombo].mat1Spectra / meanValues[sett][matCombo].mat2Spectra).mean() for
                          matCombo in settMatCombos]
        [scatterAx3.scatter(x, y, label=f'{matCombo[0].name}/{matCombo[1].name}') for x, y, matCombo in zip(scatterPointsX, scatterPointsY, settMatCombos)]
        x = np.array([1, max(scatterPointsX + scatterPointsY) * 1.05])
        scatterAx3.plot(x, x, label='1:1')
        scatterAx3.legend()

        fig4, scatterAx2 = comparisonDock.subplots(sett)  # A scatter plot of the theoretical vs observed reflectance ratio.
        scatterAx2.set_ylabel("Theoretical Ratio")
        scatterAx2.set_xlabel("Observed Ratio after Subtraction")
        scatterPointsY = [(theoryR[matCombo[0]] / theoryR[matCombo[1]]).mean() for matCombo in settMatCombos]
        scatterPointsX = [np.nanmean((meanValues[sett][matCombo].mat1Spectra - means.I0 * means.rExtra) / (
                meanValues[sett][matCombo].mat2Spectra - means.I0 * means.rExtra)) for matCombo in
                          settMatCombos]
        [scatterAx2.scatter(x, y, label=f'{matCombo[0].name}/{matCombo[1].name}') for x, y, matCombo in
         zip(scatterPointsX, scatterPointsY, settMatCombos)]
        x = np.array([1, max(scatterPointsX + scatterPointsY) * 1.05])
        scatterAx2.plot(x, x, label='1:1')
        scatterAx2.legend()
    return figs


def _generateOneRExtraCube(combo: CubeCombo, theoryR: t_.Dict[Material, pd.Series]) -> t_.Tuple[np.ndarray, np.ndarray]:
    """Given a combination of two PwsCubes imaging different materials and the theoretical reflectance for the materials
    imaged in the two PwsCubes this function will generate an estimation of the extra reflectance in the system.

    Args:
        combo: A combo of two PwsCubes imaging two different glass-material interfaces.
        theoryR: A dictionary providing Pandas `Series` giving the theoretically expected reflectance for each glass-material
            interface represented by the two PwsCubes.

    Returns:
        The first item is the array of the estimated extra reflectance of the system, expressed as values between 0 and
        1 (this is the same shape as the PwsCubes supplied to the function.) The second item is an estimate of how reliable
        each element in the array is.

    """
    data1 = combo.data1.data
    data2 = combo.data2.data
    T1 = np.array(theoryR[combo.mat1])[np.newaxis, np.newaxis, :]
    T2 = np.array(theoryR[combo.mat2])[np.newaxis, np.newaxis, :]
    denominator = data1 - data2
    nominator = T1 * data2 - T2 * data1
    arr = nominator / denominator

    #Even when a weight is calculated as zero if we have weird values (np.nan, np.inf) in arr we will get a messed up end result.
    arr[np.isinf(arr)] = 0
    arr[np.isnan(arr)] = 0
    arr[arr < 0] = 0
    arr[arr > 1] = 1
    #calculate a confidence weighting for every point in the cube.
    # According to propagation of error if we assume that TheoryR has no error
    # and the data (camera counts) has a constant error of C then the error is C * sqrt((T1-T2)*data1^2 + (T2-T1)*data2^2) / (data1 - data2)^2
    # Since we are just looking for a relative measure of confidence we can ignore C. We use the `Variance weighted average'
    # (1/stddev^2)
    #Doing this calculation with noise in Theory instead of data gives us a variance of C^2 * (data1^2 + data2^2) / (data1 - data2)^2. this seems like a better equation to use. TODO Really? Why?
    weight = (data1-data2)**2 / (data1**2 + data2**2)  # The weight is the inverse of the variance. Higher weight = more reliable data.
    return arr, weight


def generateRExtraCubes(allCombos: t_.Dict[MCombo, t_.List[CubeCombo]], theoryR: t_.Dict[Material, pd.Series], numericalAperture: float) -> \
        t_.Tuple[pwsdt.ExtraReflectanceCube, t_.Dict[t_.Union[str, MCombo], np.ndarray]]:
    """Generate a series of extra reflectance cubes based on the input data.

    Args:
        allCombos: a dict of lists CubeCombos, each keyed by a 2-tuple of Materials.
        theoryR: the theoretically predicted reflectance for each material.
        numericalAperture: The numerical aperture that the PwsCubes were imaged at. The theoryR reflectances should have
            also been calculated at this NA

    Returns:
        An `ExtraReflectanceCube` object containing data from the weighted average of all measurements.
         A dictionary where the keys are material combos and the values are tuples of the weightedMean and the weight arrays.
    """
    rExtra: t_.Dict[MCombo, np.ndarray] = {}
    rExtraWeight: t_.Dict[MCombo, np.ndarray] = {}
    # Calculate weighted sum for all measurements within a certain material combo
    for matCombo, combosList in allCombos.items():
        logging.getLogger(__name__).info(f"Calculating rExtra for: {matCombo}")
        erCubes, weights = zip(*[_generateOneRExtraCube(combo, theoryR) for combo in combosList])
        weightSum = reduce(lambda x, y: x + y, weights)  # Sum of all weights
        weightedMean = reduce(lambda x, y: x + y, [cube*weight for cube, weight in zip(erCubes, weights)]) / weightSum  # Weighted mean of ER cubes
        meanWeight = weightSum / len(weights)
        rExtra[matCombo] = weightedMean
        rExtraWeight[matCombo] = meanWeight

    # Calculate weight mean accross all material combos
    erCubes = list(rExtra.values())
    weights = list(rExtraWeight.values())
    weightSum = reduce(lambda x, y: x + y, weights) # Sum of all weights
    weightedMean = reduce(lambda x, y: x + y, [cube * weight for cube, weight in zip(erCubes, weights)]) / weightSum  # Weighted mean of ER cubes
    # meanWeight = weightSum / len(weights)
    sampleCube: pwsdt.PwsCube = list(allCombos.values())[0][0].data1
    md = pwsdt.ERMetaData(sampleCube.metadata.dict, numericalAperture)
    erCube = pwsdt.ExtraReflectanceCube(weightedMean, sampleCube.wavelengths, md)
    return erCube, rExtra






"""This was an initial attempt that relied on detecting peaks in the spectra and then performing further processing
to dtermine which peaks might be associated with reflections off the top surface of a cell."""


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Contains all code used for the analysis of data acquired with the PWS system.

Submodules
------------

.. autosummary::
    :toctree: generated/

    compilation
    pws
    warnings
    dynamics

Inheritance
-------------
.. inheritance-diagram:: pwspy.analysis.pws.PWSAnalysisSettings pwspy.analysis.pws.PWSAnalysisResults pwspy.analysis.pws.PWSAnalysis pwspy.analysis.dynamics.DynamicsAnalysisSettings pwspy.analysis.dynamics.DynamicsAnalysisResults pwspy.analysis.dynamics.DynamicsAnalysis
    :parts: 1

"""
import os
from ._abstract import AbstractAnalysisSettings, AbstractAnalysis, AbstractAnalysisResults, AbstractHDFAnalysisResults
from . import pws
from . import dynamics
from . import compilation
from ._utility import ParallelRunner
# TODO settings are missing reference IDtag but they exist in the results. Results and settings both contain extra reflectance idTag, reduntant

resources = os.path.join(os.path.split(__file__)[0], '_resources')
defaultSettingsPath = os.path.join(resources, 'defaultAnalysisSettings')

__all__ = ['AbstractAnalysisSettings', 'AbstractAnalysis', 'AbstractAnalysisResults',
           'AbstractHDFAnalysisResults', 'resources', 'defaultSettingsPath', 'pws', 'dynamics', 'compilation', 'ParallelRunner']










# -*- coding: utf-8 -*-
# Copyright 2018-2021 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
This script allows the user to select a region of an PwsCube. the spectra of this
region is then averaged over the X and Y dimensions. This spectra is then saved
as a reference dataTypes with the same initial dimensions.
Can help to make a reference when you don't actually have one for some reason
"""

import pwspy.dataTypes as pwsdt
import matplotlib.pyplot as plt
import numpy as np
from examples import PWSImagePath

plt.ion()
a = pwsdt.Acquisition(PWSImagePath).pws.toDataClass()  # Load a measurement from file.

roi = a.selectLassoROI()  # Prompt the user for a hand-drawn ROI
spec, std = a.getMeanSpectra(mask=roi)  # Get the average spectra within the ROI
newData = np.zeros(a.data.shape)
newData[:, :, :] = spec[np.newaxis, np.newaxis, :]  # Extend the averaged spectrum along the full dimensions of the original measurement.
ref = pwsdt.PwsCube(newData, a.metadata)  # Create a new synthetic measurement using the averaged spectrum
plt.plot(a.wavelengths, spec)





# -*- coding: utf-8 -*-
# Copyright 2018-2021 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
This script is based on a matlab script written by Lusik Cherkezyan for NC.
Nano uses this method to extract rms from phantom make from ChromEM cells embedded in resin.
The phantom has a strong thin-film spectrum. This script is meant to filter out the thin film components
of the fourier transfrom and extract RMS from what is left.
"""
from pwspy.dataTypes import CameraCorrection, Acquisition, Roi, PwsCube, KCube
import matplotlib.pyplot as plt
import scipy.signal as sps
import os
import numpy as np

'''User Input'''
path = r'2_7_2019 11.07'
refName = 'Cell999'  # This is an PwsCube of glass, used for normalization.
cellNames = ['Cell1', 'Cell2']  # , 'Cell3', 'Cell4','Cell5']
maskSuffix = 'resin'

# identify the depth in um to which the OPD spectra need to be integrated
integrationDepth = 2.0  ##  in um
isHannWindow = True  # Should Hann windowing be applied to eliminate edge artifacts?
subtractResinOpd = True
resetResinMasks = False
wvStart = 510  # start wavelength for poly subtraction
wvEnd = 690  # end wavelength for poly subtraction
sampleRI = 1.545  # The refractive index of the resin. This is taken from matlab code, I don't know if it's correct.
orderPolyFit = 0
wv_step = 2
correction = CameraCorrection(2000, (0.977241216, 1.73E-06, 1.70E-11))

'''************'''

b, a = sps.butter(6, 0.1 * wv_step)
opdIntegralEnd = integrationDepth * 2 * sampleRI  # We need to convert from our desired depth into an opd value. There are some questions about having a 2 here but that's how it is in the matlab code so I'm keeping it.

### load and save mirror or glass image cube
ref = PwsCube.fromMetadata(Acquisition(os.path.join(path, refName)).pws)
ref.correctCameraEffects(correction)
ref.filterDust(6, pixelSize=1)
ref.normalizeByExposure()

if subtractResinOpd:
    ### load and save reference empty resin image cube
    fig, ax = plt.subplots()
    resinOpds = {}
    for cellName in cellNames:
        resin = PwsCube.fromMetadata(Acquisition(os.path.join(path, cellName)).pws)
        resin.correctCameraEffects(correction)
        resin.normalizeByExposure()
        resin /= ref
        resin = KCube.fromPwsCube(resin)
        if resetResinMasks:
            [resin.metadata.acquisitionDirectory.deleteRoi(name, num) for name, num, fformat in resin.metadata.acquisitionDirectory.getRois() if name == maskSuffix]
        if maskSuffix in [name for name, number, fformat in resin.metadata.acquisitionDirectory.getRois()]:
            resinRoi = resin.metadata.acquisitionDirectory.loadRoi(maskSuffix, 1)
        else:
            print('Select a region containing only resin.')
            resinRoi = resin.selectLassoROI()
            resin.metadata.acquisitionDirectory.saveRoi(maskSuffix, 1, resinRoi)
        resin.data -= resin.data.mean(axis=2)[:, :, np.newaxis]
        opdResin, xvals = resin.getOpd(isHannWindow, indexOpdStop=None, mask=resinRoi.mask)
        resinOpds[cellName] = opdResin
        ax.plot(xvals, opdResin, label=cellName)
        ax.vlines([opdIntegralEnd], ymin=opdResin.min(), ymax=opdResin.max())
        ax.set_xlabel('OPD')
        ax.set_ylabel("Amplitude")
    ax.legend()
    plt.pause(0.2)

rmses = {}  # Store the rms maps for later saving
for cellName in cellNames:
    cube = PwsCube.fromMetadata(Acquisition(os.path.join(path, cellName)).pws)
    cube.correctCameraEffects(correction)
    cube.normalizeByExposure()
    cube /= ref
    cube.data = sps.filtfilt(b, a, cube.data, axis=2)
    cube = KCube.fromPwsCube(cube)

    ## -- Polynomial Fit
    print("Subtracting Polynomial")
    polydata = cube.data.reshape((cube.data.shape[0] * cube.data.shape[1], cube.data.shape[2]))
    polydata = np.rollaxis(polydata, 1)  # Flatten the array to 2d and put the wavenumber axis first.
    cubePoly = np.zeros(polydata.shape)  # make an empty array to hold the fit values.
    polydata = np.polyfit(cube.wavenumbers, polydata,
                          orderPolyFit)  # At this point polydata goes from holding the cube data to holding the polynomial values for each pixel. still 2d.
    for i in range(orderPolyFit + 1):
        cubePoly += (np.array(cube.wavenumbers)[:, np.newaxis] ** i) * polydata[i,
                                                                       :]  # Populate cubePoly with the fit values.
    cubePoly = np.moveaxis(cubePoly, 0, 1)
    cubePoly = cubePoly.reshape(cube.data.shape)  # reshape back to a cube.
    # Remove the polynomial fit from filtered cubeCell.
    cube.data = cube.data - cubePoly

    rmsData = np.sqrt(np.mean(cube.data ** 2, axis=2)) #This can be compared to rmsOPDIntData, when the integralStopIdx is high and we don't do spectral subtraction they should be equivalent.

    # Find the fft for each signal in the desired wavelength range
    opdData, xvals = cube.getOpd(isHannWindow, None)

    if subtractResinOpd:
        opdData = opdData - resinOpds[cellName]

    try:
        integralStopIdx = np.where(xvals >= opdIntegralEnd)[0][0]
    except IndexError:  # If we get an index error here then our opdIntegralEnd is probably bigger than we can achieve. Just use the biggest value we have.
        integralStopIdx = None
        opdIntegralEnd = max(xvals)
        print(f'Integrating to OPD {opdIntegralEnd}')

    opdSquared = np.sum(opdData[:, :, :integralStopIdx] ** 2,axis=2)  # Parseval's theorem tells us that this is equivalent to the sum of the squares of our original signal
    opdSquared *= len(cube.wavenumbers) / opdData.shape[2]  # If the original data and opd were of the same length then the above line would be correct. Since the fft has been upsampled. we need to normalize.
    rmsOpdIntData = np.sqrt(opdSquared)  # this should be equivalent to normal RMS if our stop index is high and resin subtraction is disabled.

    cmap = plt.get_cmap('jet')
    fig, axs = plt.subplots(1, 2, sharex=True, sharey=True)
    im = axs[0].imshow(rmsData, cmap=cmap, clim=[np.percentile(rmsData, 0.5), np.percentile(rmsData, 99.5)])
    fig.colorbar(im, ax=axs[0])
    axs[0].set_title('RMS')
    im = axs[1].imshow(rmsOpdIntData, cmap=cmap,
                       clim=[np.percentile(rmsOpdIntData, 0.5), np.percentile(rmsOpdIntData, 99.5)])
    fig.colorbar(im, ax=axs[1])
    axs[1].set_title(f'RMS from OPD below {opdIntegralEnd} after resin OPD subtraction')
    fig.suptitle(cellName)
    rmses[cellName] = rmsOpdIntData
    plt.pause(0.2)
plt.pause(0.5)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
from dataclasses import dataclass
from typing import Tuple, List

import numpy as np

from ._abstract import AbstractCompilerSettings, AbstractRoiCompilationResults, AbstractRoiCompiler
from .. import warnings
from ..dynamics import DynamicsAnalysisResults
from ...dataTypes import Roi
from ...dataTypes._other import RoiFile


@dataclass
class DynamicsCompilerSettings(AbstractCompilerSettings):
    """These settings determine how a Dynamics acquisition should be compiled."""
    meanReflectance: bool = False
    rms_t_squared: bool = False
    diffusion: bool = False


@dataclass
class DynamicsRoiCompilationResults(AbstractRoiCompilationResults):
    cellIdTag: str
    analysisName: str
    reflectance: float
    rms_t_squared: float
    diffusion: float


class DynamicsRoiCompiler(AbstractRoiCompiler):
    def __init__(self, settings: DynamicsCompilerSettings):
        super().__init__(settings)

    def run(self, results: DynamicsAnalysisResults, roi: Roi) -> Tuple[DynamicsRoiCompilationResults, List[warnings.AnalysisWarning]]:
        reflectance = self._avgOverRoi(roi, results.meanReflectance) if self.settings.meanReflectance else None
        rms_t_squared = self._avgOverRoi(roi, results.rms_t_squared) if self.settings.rms_t_squared else None  # Unlike with diffusion we should not have any nan values for rms_t. If we get nan then something is wrong with the analysis.
        diffusion = self._avgOverRoi(roi, results.diffusion, np.logical_not(np.isnan(results.diffusion))) if self.settings.diffusion else None  # Don't include nan values in the average. Diffusion is expected to have many Nans due to low SNR.

        results = DynamicsRoiCompilationResults(
                    cellIdTag=results.imCubeIdTag,
                    analysisName=results.analysisName,
                    reflectance=reflectance,
                    rms_t_squared=rms_t_squared,
                    diffusion=diffusion)
        warns = []  # Strip None from warns list
        return results, warns

    @staticmethod
    def _avgOverRoi(roi: Roi, arr: np.ndarray, condition: np.ndarray = None) -> float:
        """Returns the average of arr over the ROI.
        if condition is provided then only value of arr where the condition is satisfied are included."""
        assert len(arr.shape) == 2
        if condition is not None:
            return arr[np.logical_and(roi.mask, condition)].mean()
        else:
            return arr[roi.mask].mean()




# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from pwspy.dataTypes import DynCube
import numpy as np
import os
import pandas as pd

# This is very nonfunctional. just a sketch

wDir = r''
cellNums = range(1001, 1015)
background = 1997  # Temporal Background for noise subtraction
bwName = 'nuc'
n_medium = 1.37  # RI of the media (avg RI of chromatin)

sigmaCells = range(1, 15)
sigmaName = 'p0'

fileName = f'{bwName}_Autocorr'
dataAutocorr = {}
dataDiff = []

output = []

bgCube = DynCube.loadAny(os.path.join(wDir, f'Cell{background}'))
backgroundACF = bgCube.getAutocorr() #The background ACF

for cellNum in cellNums:
    cube = DynCube.loadAny(os.path.join(wDir, f'Cell{cellNum}'))
    rois = cube.metadata.getRois()
    for roi in rois:
        an = cube.loadAnalysis()
        ac = cube.getAutocorrelation()
        bLim = backgroundACF[:,:,0][roi.mask].mean() #The mean background ACF0 value for the ROI
        rmsT_sq = (ac[:,:,0] - bLim).mean()  # background subtracted sigma_t^2

        # Remove pixels with low SNR. default threshold removes values where first ACF point is less than sqrt(2) of background ACF
        ac = ac[ac[:, :, 0] > np.sqrt(2)*bLim[:,:,None]]

        normBsCorr = ac - backgroundACF[roi.mask[:,:,None]]  # Background Subtraction
        normBsCorr = normBsCorr / np.abs(normBsCorr[:, 0])[:, None]  # Normalization

        # Remove negative values before taking log
        normBsCorr[normBsCorr < 0] = np.nan

        dt = cube.times[1] - cube.times[0]
        k = (n_medium * 2 * np.pi) / cube.metadata.wavelength
        dSlope = np.diff(np.log(normBsCorr).mean(axis=(0, 1))) / (dt*4*k**2)
        dSlope = dSlope[0]

        output.append({'Cell #': f'{cellNum}_{roi.name}{roi.number}',
         'D': dSlope,
         'Sigma_t^2 (b sub)': rmsT_sq,
         'Sigma_s': an.rms[roi.mask].mean(),
         'Reflectance': an.reflectance[roi.mask].mean()})

output = pd.DataFrame(output)






# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""Useful subpackages ranging many different topics

Modules
----------
.. autosummary::
   :toctree: generated/

   acquisition
   DConversion
   fileIO
   fluorescence
   machineVision
   micromanager
   misc
   plotting
   reflection


"""

import os

thinFilmPath = os.path.join(os.path.split(__file__)[0], 'thinFilmInterferenceFiles')

__all__ = ['acquisition', 'DConversion', 'fileIO', 'fluorescence', 'machineVision', 'misc',
           'micromanager', 'plotting', 'reflection']



# Copyright 2018-2021 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
This script loads analysis results from a list of PWS acquisitions and uses the "PWSRoiCompiler" class to get average output
values within the ROIs for each acquisition. The compiled results are then placed into a Pandas dataframe.
"""
import pandas
from pwspy.analysis.compilation import PWSRoiCompiler, PWSCompilerSettings
from examples import PWSExperimentPath
import pwspy.dataTypes as pwsdt

tmpList = []
compiler = PWSRoiCompiler(PWSCompilerSettings(reflectance=True, rms=True))

listOfAcquisitions = [pwsdt.Acquisition(i) for i in PWSExperimentPath.glob("Cell[0-9]")]
for acquisition in listOfAcquisitions:
    for analysisName in acquisition.pws.getAnalyses():
        analysisResults = acquisition.pws.loadAnalysis(analysisName)
        for roiSpec in acquisition.getRois():
            roiFile = acquisition.loadRoi(*roiSpec)
            results, warnings = compiler.run(analysisResults, roiFile.getRoi())

            if len(warnings) > 0:
                print(warnings)

            tmpList.append(dict(
                acquisition=acquisition,
                cellNumber=acquisition.getNumber(),
                analysisResults=analysisResults,
                rms=results.rms,
                reflectance=results.reflectance,
                roiNum=roiFile.number,
                roiName=roiFile.name
            ))

dataFrame = pandas.DataFrame(tmpList)
print(dataFrame)
